/* X_events.c : part of FancyLauncher
 *
 * provides: X events and loop
 *
 * author: Alexandre David
 * e-mail: adavid@docs.uu.se
 * http://www.docs.uu.se/~adavid
 * license: GPL
 */


#include <X11/Xlib.h>
#include <X11/Xresource.h>
#include <X11/Xutil.h>
#include <Imlib.h>
#include <sys/utsname.h>
#include <X11/keysym.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <values.h>
#include "debug.h"
#include "utils.h"
#include "gadget.h"
#include "pixmaps.h"
#include "Xcore.h"
#include "codes.h"
#include "font.h"
#include "tooltips.h"
#include "load_save.h"

extern int main_argc;
extern char **main_argv;

XContext xid_context=0;
Atom  wmDeleteWindow;


TimedPeriodicAction *jobs=NULL;
int n_jobs=0;
double tooltip_period=2.0;
double tooltip_time=0.0;


#if DEBUG
static char *evTable[]={
  "0x0","0x1",
  "KeyPress",
  "KeyRelease",
  "ButtonPress",
  "ButtonRelease",
  "MotionNotify",
  "EnterNotify",
  "LeaveNotify",
  "FocusIn",
  "FocusOut",
  "KeymapNotify",
  "Expose",
  "GraphicsExpose",
  "NoExpose",
  "VisibilityNotify",
  "CreateNotify",
  "DestroyNotify",
  "UnmapNotify",
  "MapNotify",
  "MapRequest",
  "ReparentNotify",
  "ConfigureNotify",
  "ConfigureRequest",
  "GravityNotify",
  "ResizeRequest",
  "CirculateNotify",
  "CirculateRequest",
  "PropertyNotify",
  "SelectionClear",
  "SelectionRequest",
  "SelectionNotify",
  "ColormapNotify",
  "ClientMessage",
  "MappingNotify",
  "LASTEvent"
};
#endif



/* register a periodic gadget
 */
void registerPeriodicGadget(Gadget *gad) {
  int i;

  pushArgs(8,(BUF,"%s",PTR2STR(gad)));
  assert(gad==NULL || IS_PERIODIC(gad),"Periodic gadget expected");
  assert((jobs!=NULL) || (n_jobs==0),"Invalid data");

  for(i=0;i<n_jobs;i++)
    if (jobs[i].gad==gad)
      return; /* already registered */

  n_jobs++;
  trace( jobs=(TimedPeriodicAction*)safeAllocPtr(ralloc(jobs,n_jobs*sizeof(TimedPeriodicAction))) );
  jobs[n_jobs-1].time=(gad ? 0.0 : tooltip_time );
  jobs[n_jobs-1].gad=gad;
  /* debugPrint((GREEN"New job registered for: %g"BLACK"\n",jobs[n_jobs-1].time)); */
}


/* unregister a periodic gadget
 */
void unregisterPeriodicGadget(Gadget *gad) {
  int i;

  pushArgs(8,(BUF,"%s",PTR2STR(gad)));
  assert(gad==NULL || IS_PERIODIC(gad),"Periodic gadget expected");
  assert((jobs!=NULL) || (n_jobs==0),"Invalid data");

  for(i=n_jobs-1;i>=0;i--)
    if (jobs[i].gad==gad) {
      memcpy(jobs,jobs+1,(n_jobs-i-1)*sizeof(TimedPeriodicAction));
      n_jobs--;
      /* debugPrint((GREEN"Unregistered job"BLACK"\n")); */
      return;
    }
  /* debugPrint(("gadget not found, could not unregister!\n")); */
}


/* create the window for a button
 * original code taken from epplet.c
 */
static void createWindow(Gadget *button,int index) {
  XSetWindowAttributes attr;
  Atom                a;
  XTextProperty       xtp;
  XClassHint         *xch;
  XSizeHints          sh;
  struct utsname      ubuf;
  MWMHints            mwm;
  unsigned long       val;
  Display            *disp;
  Window root;

  assert(button!=NULL,"Button argument null!");
  assert(index>=0,"Invalid index");
  assert(xid_context!=0,"Context is not initialized!");

  disp=display;

  if (!disp) {
    fprintf(stderr, "FancyLauncher Error: display not opened!\n");
    exit(1);
  }
  XSelectInput(disp, DefaultRootWindow(disp), PropertyChangeMask);

  /* create a window with everything set */
  attr.backing_store = NotUseful;
  attr.override_redirect = False;
  attr.colormap = Imlib_get_colormap(im_data);
  attr.border_pixel = 0;
  attr.background_pixel = 0;
  attr.save_under = False;
  attr.event_mask = StructureNotifyMask | ButtonPressMask |
    ButtonReleaseMask | EnterWindowMask | ButtonMotionMask |
    LeaveWindowMask | KeyPressMask | KeyReleaseMask | PointerMotionMask;
  root=DefaultRootWindow(disp);
  button->win = XCreateWindow(disp, root,
			      button->x, button->y, button->width, button->height, 0,
			      im_data->x.depth, InputOutput,
			      Imlib_get_visual(im_data),
			      CWOverrideRedirect | CWSaveUnder |
			      CWBackingStore | CWColormap | CWBackPixel |
			      CWBorderPixel | CWEventMask, &attr);
  XSetTransientForHint(disp, button->win, root);
  debugPrint(("Window %d created ... ",index));
  
  /* set hints to be borderless */
  mwm.flags = MWM_HINTS_DECORATIONS;
  mwm.functions = 0;
  mwm.decorations = 0;
  mwm.inputMode = 0;
  mwm.status = 0;
  a = XInternAtom(disp, "_MOTIF_WM_HINTS", False);
  XChangeProperty(disp, button->win, a, a, 32, PropModeReplace,
		  (unsigned char *)&mwm, sizeof(MWMHints) / 4);
  debugPrint(("borderless set ... "));
  
  /* set the window title , name , class */
  XStoreName(disp, button->win, "FancyLauncher");
  xch = XAllocClassHint();
  xch->res_name = "FancyLauncher";
  xch->res_class = "FancyLauncher";
  XSetClassHint(disp, button->win, xch);
  XFree(xch);
  debugPrint(("title set ... "));

  /* set the size hints */
  sh.flags = PSize | PMinSize | PMaxSize | PPosition;
  sh.x = button->x;
  sh.y = button->y;
  sh.width = button->width;
  sh.height = button->height;
  sh.min_width = 16;
  sh.min_height = 16;
  sh.max_width = 1024;
  sh.max_height = 768;
  XSetWMNormalHints(disp, button->win, &sh);
  debugPrint(("size set ... "));
  
  /* set the command hint */
  if (index==0 && PAGEOF(button)==0) {
    XSetCommand(disp, button->win, main_argv, main_argc);
    debugPrint(("command hint set ... "));
  }

  /* set the client machine name */
  if (!uname(&ubuf)) {
    xtp.encoding = XA_STRING;
    xtp.format = 8;
    xtp.value = (unsigned char *)ubuf.nodename;
    xtp.nitems = strlen((char *)(xtp.value));
    XSetWMClientMachine(disp, button->win, &xtp);
    debugPrint(("machine: %s set ...",ubuf.nodename));
  }
  
  /* set the icons name property */
  XSetIconName(disp, button->win, "FancyLauncher");
  debugPrint(("icon name ... "));
  
  /* set sticky & arrange ignore */
  if (IS_STICKY(button)) {
    val = (1 << 0) /* | (1 << 9) */ ;
    a = XInternAtom(disp, "_WIN_STATE", False);
    XChangeProperty(disp, button->win, a, XA_CARDINAL, 32, PropModeReplace,
		    (unsigned char *)&val, 1);
    debugPrint(("sticky ... "));
  }
  
  /* set the default layer to below ; or if <0 to above */
  if (IS_BELOW(button)) {
    val = IS_BELOW(button)>1 ? 255 : 0 ;
    a = XInternAtom(disp, "_WIN_LAYER", False);
    XChangeProperty(disp, button->win, a, XA_CARDINAL, 32, PropModeReplace,
		    (unsigned char *)&val, 1);
    debugPrint(("below ... "));
  }
  
  /* set skip focus, skip winlist dont cover skip taskbar flags */
  val = (1 << 0) | (1 << 1) | (1 << 2) | (1 << 5);
  a = XInternAtom(disp, "_WIN_HINTS", False);
  XChangeProperty(disp, button->win, a, XA_CARDINAL, 32, PropModeReplace,
		  (unsigned char *)&val, 1);
  XSaveContext(disp,button->win,xid_context,(XPointer)button);
  XSetWMProtocols(disp,button->win,&wmDeleteWindow,1);
  debugPrint(("complete\n"));

#if DEBUG
  if (index==0) {
    int n,i;
    Atom *list=XListProperties(display,button->win,&n);
    if (list) {
      for(i=0;i<n;i++) {
	char *name=XGetAtomName(display,list[i]);
	if (name) {
	  printf("\t%s\n",XGetAtomName(display,list[i]));
	  XFree(name);
	}
      }
      XFree(list);
    }
  }
#endif
}


/* initialize the buttons and show them
 */
void showButtons(int page) {
  if (page>=0 && page<npages) {
    int i,n;
    Gadget *buttons=getButtons(page,&n);
    assert((buttons==NULL)==(n==0),"Inconsistent data");
    for(i=0;i<n;i++) {
      debugPrint(("Show button[%d].%d\n",page,i));
      if (!buttons[i].win)
	createWindow(buttons+i,i);
      assert(buttons[i].win!=0,"Window not created!");
      XMapWindow(display, buttons[i].win);
    }
  }
}

/* same for the extended gadgets
 * register the periodic gadgets as well
 */
void showExtGadgets() {
  if (nb_extGadgets>0) {
    int i;
    for(i=0;i<nb_extGadgets;i++) {
      if (!extGadgets[i].win)
	createWindow(extGadgets+i,npages);
      ANY(extGadgets+i).opt[I_PAGE]=0;
      XMapWindow(display,extGadgets[i].win);
      if (IS_PERIODIC(extGadgets+i))
	registerPeriodicGadget(extGadgets+i);
    }
  }
}

#define SNAPIFY(S) (abs(*vx-(S))<=snap) *vx=(S)


static int snapVertex(int *vx,int vw,int x,int w,int snap) {
  if SNAPIFY(x);
  else if SNAPIFY(x+w);
  else if SNAPIFY(x-vw);
  else if SNAPIFY(x+w-vw);
  else return 0;
  return 1;
}


/* check position of a gadget
 */
static int checkPosition(Gadget *gad,int posx,int posy,int snapit) {
  Window dummy;
  unsigned int rw,rh;
  int i;

  assert(gad!=NULL,"Null gadget as argument!");

  /* only rw,rh are interesting */
  XGetGeometry(display,XDefaultRootWindow(display),
	       &dummy,&i,&i,&rw,&rh,&i,&i);

  if (snapit) {
    int n;
    int snap,snappedx=0,snappedy=0;
    Gadget *all;

    snap=IS_SNAPPED(gad);
    all=getButtons(current_page,&n);
    assert((n==0)==(all==NULL),"Inconsistent data");
    i=0;
    if (all && snap) {
      for(;i<n;i++)
	if (gad!=all+i) {
	  if (!snappedx)
	    snappedx=snapVertex(&gad->x,gad->width,
				all[i].x,all[i].width,snap);
	  if (!snappedy)
	    snappedy=snapVertex(&gad->y,gad->height,
				all[i].y,all[i].height,snap);
	  if (snappedx && snappedy)
	    break;
	}
    }
    if (i==n) {
      assert((extGadgets!=NULL) || (nb_extGadgets==0),"Inconsistent data");
      for(i=0;i<nb_extGadgets;i++)
	if (gad!=extGadgets+i) {
	  if (!snappedx)
	    snappedx=snapVertex(&gad->x,gad->width,
				extGadgets[i].x,
				extGadgets[i].width,snap);
	  if (!snappedy)
	    snappedy=snapVertex(&gad->y,gad->height,
				extGadgets[i].y,
				extGadgets[i].height,snap);
	  if (snappedx && snappedy)
	    break;
	}
    }
  }

  if (posx>=0 && posx+gad->width<=rw &&
      posy>=0 && posy+gad->height<=rh)
    return 0;
  if (posx<0)
    gad->x=0;
  else if (posx+gad->width>rw)
    gad->x=rw-gad->width;
  if (posy<0)
    gad->y=0;
  else if (posy+gad->height>rh)
    gad->y=rh-gad->height;
  return 1;
}


static int checkedMove(int *dx,int *dy,Gadget *gad,int nb) {
  Window dummy;
  int i,xx=*dx,yy=*dy;
  unsigned int rw,rh;

  XGetGeometry(display,XDefaultRootWindow(display),
	       &dummy,&i,&i,&rw,&rh,&i,&i);

  for(i=0;i<nb;i++) {
    if (gad[i].x+xx<0) {
      if (-xx>gad[i].x)
	xx=-gad[i].x;
    } else if (gad[i].x+gad[i].width+xx>=rw) {
      if (xx>rw-gad[i].x-gad[i].width)
	xx=rw-gad[i].x-gad[i].width;
    }
    if (gad[i].y+yy<0) {
      if (-yy>gad[i].y)
	yy=-gad[i].y;
    } else if (gad[i].y+gad[i].height+yy>=rh) {
      if (yy>rh-gad[i].y-gad[i].height)
	yy=rh-gad[i].y-gad[i].height;
    }
  }

  if (xx==0 && yy==0)
    return 0;
  *dx=xx;
  *dy=yy;
  return 1;
}


static void setPage(int p) {
  if (p>=0 && p<npages && p!=current_page) {
    int i,n1,n2,n;

    Gadget *old_buttons=getButtons(current_page,&n1);
    Gadget *new_buttons=getButtons(p,&n2);
    n=( n1 > n2 ? n1 : n2 );
    current_page=p;

    assert(old_buttons!=NULL,"No old buttons!");
    assert(new_buttons!=NULL,"No new buttons!");

    if (n==0) /* old and new are empty! */
      return;
    debugPrint(("setPage(%d)\n",p));

    for(i=0;i<n;i++) {
      debugPrint(("Change button[%d].%d ",p,i));
      if (i>=n2) { /* fewer buttons in the new page */
	assert(n1>n2,"Violated: more new than old buttons");
	XUnmapWindow(display, old_buttons[i].win);
	IS_MAPPED(old_buttons+i)=0;
	debugPrint(("unmapped old window\n"));
      } else if (i>=n1) { /* more buttons in the new page */
	XEvent ev;
	assert(n2>n1,"Violated: more old than new buttons");
	if (!new_buttons[i].win) {
	  int j;
	  for(j=0;j<npages;j++)
	    if (i<NBUTTONS(j) && pages[j].buttons[i].win) {
	      new_buttons[i].win=pages[j].buttons[i].win;
	      debugPrint(("Found window[%d] in page %d\n",i,j));
	      break;
	    }
	  assert((j==npages)==(new_buttons[i].win==0),"Wrong algo!\n");
	  if (j==npages) {
	    createWindow(new_buttons+i,i);
	    debugPrint(("created new window, "));
	  }
	}
	XMapWindow(display, new_buttons[i].win);
	XMaskEvent(display, StructureNotifyMask, &ev);
	debugPrint(("mapped new window\n"));
      } else { /* recycle */
	assert(i<n1 && i<n2,"Index out of range");
	assert(old_buttons[i].win!=0,"Old window was null!");
	new_buttons[i].win=old_buttons[i].win;
	IS_MAPPED(new_buttons+i)=1;
	debugPrint(("recycled window\n"));
      }
      if (i<n2) {
	ImlibData *id=im_data;
	assert(new_buttons[i].win!=0,"New window is null!");
	/* to the right button first ! */
	XSaveContext(display,new_buttons[i].win,xid_context,(XPointer)(new_buttons+i));
	/* 2 calls because I want it work with twm
	 * with the limit of the placement activated.
	 * XMoveResizeWindow causes troubles if I use
	 * a window in a corner that changes size in another
	 * page ; always trouble anyway :(
	 */
	XResizeWindow(display,new_buttons[i].win,
		      new_buttons[i].width,new_buttons[i].height);
	XSync(display,True); /* discard the resize, otherwise race with move */
	XMoveWindow(display,new_buttons[i].win,
		    new_buttons[i].x,new_buttons[i].y);
	/* reset the mask otherwise, get into trouble */
	XShapeCombineMask(display,new_buttons[i].win,ShapeBounding,
			  0,0,0,ShapeSet);
	XSync(display,False);
	paintInitButton(id,new_buttons+i);
      }
    }
  }
}


void exec_command(char *command,Gadget *button) {
  int i=0;
  pid_t pid;
  debugPrint(("action: %s\n",command));

  if (command) {
    int token=readToken(command,&i);
    if (button==NULL && token!=EXEC && token!=WAIT && token!=PAGE) {
      fprintf(stderr,"Command not valid in this context: %s\n",command);
      return;
    }
    switch(token) {
    case EXEC:
    case WAIT:
      switch(pid=fork()) {
      case -1:
	perror("fork");
	break;
      case 0:
	execl("/bin/sh", "/bin/sh", "-c", command+i, NULL);
	perror(command+i);
	exit(1);
      default:
	if (token==WAIT) {
	  waitpid(pid,NULL,WUNTRACED);
	  if (button!=NULL) {
	    /* Imlib_kill_image(im_data,getButtonPixmap(im_data,button)); */
	    paintReInitButton(im_data,button);
	  }
	}
      }
      break;

    case EXIT:
      saveButtons();
      exit(0);

    case CONFIGURE:
      hide_tooltip();
      usleep(100000);
      show_tooltip(button,"Configure: function not yet implemented!",0);
      debugPrint(("Configure later on: %s\n",tooltip_win?"ok":"not ok"));
      break;

    case ABOUT:
      hide_tooltip();
      usleep(100000);
      show_tooltip(button,
		   "****  FancyLauncher ****\n\n"
		   " version 0.10, July 2000\n\n"
		   "   by Alexandre David\n"
		   "    adavid@docs.uu.se\n"
		   " http://www.docs.uu.se",0);
      break;

    case NEXT_PAGE:
      setPage(current_page+1);
      break;

    case PREVIOUS_PAGE:
      setPage(current_page-1);
      break;

    case PAGE:
      if (jumpBlank(command,&i)) {
	int new_page=atoi(command+i)-1;
	debugOK();
	if (button)
	  setPage(new_page);
	else /* pospone setPage */
	  current_page=new_page;
	break;
      }

    default:
      fprintf(stderr,"Invalid command: %s\n",command);
    }
  }
}


double get_time() {
  struct timeval tv;
  gettimeofday(&tv,NULL);
  return (double)tv.tv_sec + ((double)tv.tv_usec)/1000000.0;
}


/* execute jobs if necessary,
 * update the delays for the next time,
 * returns the updated delay to wait
 */
struct timeval* get_delay(double time,struct timeval *tv,
			  Gadget *gad,int *nfd,fd_set *fdset) {
  int i,j;
  double now=get_time();

  assert((jobs!=NULL) || (n_jobs==0),"Inconsistent data");

  if (n_jobs==0)
    return NULL;

  for(i=0,j=0;i<n_jobs;i++) {
    assert(jobs[i].gad==NULL || IS_PERIODIC(jobs[i].gad),"Invalid gadget type");
    if (jobs[i].time<=time) {
      if (jobs[i].gad) {
	jobs[i].time=now+PERIODIC(jobs[i].gad).updateGadget(jobs[i].gad,nfd,fdset,0);
      } else {
	show_tooltip(gad,NULL,1);
	jobs[i].time=now+(gad&&IS_PERIODIC(gad)?
			  PERIODIC(gad).updateGadget(gad,nfd,fdset,1):1e6);
      }
    } else if (jobs[i].gad)
      PERIODIC(jobs[i].gad).addFD(jobs[i].gad,nfd,fdset);
    if (jobs[i].time<jobs[j].time)
      j=i;
  }
  
  assert(tv!=NULL,"Null tval struct passed");
  /* "now" is now the min delay to wait from now */
  now=jobs[j].time-now;
  if (now<0.0) { /* Yes, this happens! */
    /* fprintf(stderr,"Substraction error\n"); */
    now=-now;
  }
  tv->tv_sec=(long)now;
  if (tv->tv_sec<0)
    return NULL;
  tv->tv_usec=(long)((now-(double)tv->tv_sec)*1000000.0);
  if (tv->tv_usec<0) {
    if (tv->tv_sec==0)
      tv->tv_sec=1;
    tv->tv_usec=0;
  }
  return tv;
}


/* X loop
 */
void X_loop() {
  XEvent *ev;
  Gadget *gad=NULL;
  double time;
  int n_ev=8,skip_update=0,xfd,posx=0,posy=0;
  int skip_moves=0;
  int pressed_x=0,pressed_y=0,pressed_b=0;
  struct timeval tval,*delay;
  fd_set fdset;
  int button_state=0;
  Gadget *focused_gadget=NULL;

  assert(im_data!=NULL,"Imlib data is null!");
  assert(display,"Display not initialized!");

  xfd=ConnectionNumber(display);
  ev=(XEvent*)safeAllocPtr(malloc(8*sizeof(XEvent)));

  for(;;) {
    int i,n=0;

    time=get_time();
    XFlush(display);

    /* collect all events */
    for(i=0;XPending(display);i++,n++) {
      assert(i<=n_ev,"Event queue overflow");
      if (i==n_ev) {
	n_ev+=XPending(display);
	ev=(XEvent*)safeAllocPtr(realloc(ev,n_ev*sizeof(XEvent)));
	debugPrint((YELLOW"event queue: %d\n"BLACK,n_ev));
      }
      XNextEvent(display,ev+i);
      if (skip_moves && ev[i].type==ConfigureNotify)
	ev[i].type=0;
    }

    /* simple prune events */
    for(i=n-1;i>0;i--) {
      int j;      
      for(j=i-1;j>=0;j--)
	if (ev[j].type==ev[i].type &&
	    ev[j].xany.window==ev[i].xany.window)
	  ev[j].type=0;
    }

    if (n) /* reset tooltip wait */
      trace(unregisterPeriodicGadget(NULL));

    /* handle events */
    for(i=0;i<n;i++) {
      XEvent *xev=&ev[i];
      int j;

      assert(xev->type>=0,"Invalid type");
      assert(xev->type<sizeof(evTable)/sizeof(char*),"Invalid type");
      debugPrint(("XEvent: %s\n",evTable[xev->type]));

      if (xev->type>0 &&
	  XFindContext(display,
		       xev->xany.window,
		       xid_context,
		       (XPointer*)&gad) != XCNOENT
	  && gad!=NULL
	  && (!IS_BUTTON(gad) || PAGEOF(gad)==current_page)) {

	/* button : the button where the event occured! */
	assert(xev->xany.window==gad->win,"Bad gadget found");

	if (IS_BUTTON(gad)) {
	  switch(xev->type) {
	  
	  case MapNotify:
	    XMoveWindow(display,gad->win,gad->x,gad->y);
	    IS_MAPPED(gad)=1;
	    paintInitButton(im_data,gad);
	    button_state=0;
	    break;
	  

	  case ConfigureNotify:
	    debugPrint(("\t(x,y) = %d,%d ; widthxheight = %dx%d\n",
			xev->xconfigure.x,xev->xconfigure.y,
			xev->xconfigure.width,xev->xconfigure.height));
	    if (IS_MAPPED(gad) && !skip_moves) {
	      Window dummy;
	      int posx=gad->x,posy=gad->y;
	      /* when resize is sent, x,y=0,0 !!!
		 wrong:
		 button->x=xev->xconfigure.x;
		 button->y=xev->xconfigure.y;
		 right: compute the position
	      */
	      XTranslateCoordinates(display,gad->win,
				    XDefaultRootWindow(display),
				    0,0,&gad->x,&gad->y,&dummy);
	      j=(gad->width^xev->xconfigure.width) |
		(gad->height^xev->xconfigure.height);
	      gad->width=xev->xconfigure.width;
	      gad->height=xev->xconfigure.height;
	      debugPrint(("\t(x,y) = %d,%d ; widthxheight = %dx%d\n",
			  gad->x,gad->y,
			  gad->width,gad->height));
	      if (checkPosition(gad,gad->x,gad->y,0)) {
		XMoveWindow(display,gad->win,gad->x,gad->y);
		break;
	      }
	      if (!skip_update && (j | (posx^gad->x) | (posy^gad->y))) {
		(j?paintReInitButton:paintInitButton)(im_data,gad);
		button_state=0;
	      }
	    }
	    break;
	  

	  case ButtonPress:
	    debugPrint(("\t(x,y) = %d,%d ; button = %d ; state = %d\n",
			xev->xbutton.x,xev->xbutton.y,
			xev->xbutton.button,
			xev->xbutton.state));
	  
	    pressed_x=xev->xbutton.x;
	    pressed_y=xev->xbutton.y;
	    pressed_b=xev->xbutton.button;
	    posx=gad->x;
	    posy=gad->y;
	    paintButton(im_data,gad,PRESSED);
	    button_state=-1;
	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();
	    break;
	  
	    
	  case EnterNotify:
	    paintButton(im_data,gad,FOCUSED);
	    button_state=1;
	    if (SHOWS_TOOLTIP(gad)) {
	      debugPrint((GREEN"Activated tooltip"BLACK"\n"));
	      focused_gadget=gad;
	    }
	    break;
	    

	  case ButtonRelease:
	    debug( if (xev->type==ButtonRelease)
		   printf("\t(x,y) = %d,%d ; button = %d ; state = %d\n",
			  xev->xbutton.x,xev->xbutton.y,
			  xev->xbutton.button,
			  xev->xbutton.state) );
	    
	    skip_moves=0;
	    if (skip_update) {
	      int nb;
	      Gadget *all;
	    
	      assert(current_page>=0 && current_page<npages,"Page overflow");
	      all=getButtons(current_page,&nb);
	      assert(all!=NULL,"No button found");
	      skip_update=0;

	      for(j=0;j<nb;j++)
		paintInitButton(im_data,all+j);
	    } else
	      paintInitButton(im_data,gad);
	    XFlush(display);
	    paintButton(im_data,gad,FOCUSED);
	    button_state=1;
	    j=current_page;
	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();
	    XFlush(display);
	    if ((xev->xbutton.state&5)==0 &&
		xev->xbutton.x>=0 && xev->xbutton.x<gad->width &&
		xev->xbutton.y>=0 && xev->xbutton.y<gad->height)
	      exec_command(getAction(gad,xev->xbutton.button-1),gad);
	    if (j!=current_page) {
	      button_state=0;
	      focused_gadget=NULL;  
	    }
	    break;
	  

	  case LeaveNotify:
	    paintButton(im_data,gad,NORMAL);
	    button_state=0;
	    focused_gadget=NULL;
	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();
	    break;
	  

	  case MotionNotify: /* only drag, according to the event mask */
	    debugPrint(("\t(x,y) = %d,%d ; state = %d\n",
			xev->xmotion.x,xev->xmotion.y,xev->xmotion.state));

	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();

	    if (xev->xmotion.state&(256|512|1024)) {
	      skip_moves=1;
	      if (xev->xmotion.state&4) {
		int dx,dy;
		
		debugPrint(("\t+(%d,%d)\n",dx,dy));
		
		if (pressed_b==1) {
		  
		  dx=xev->xmotion.x-pressed_x+gad->x-posx;
		  dy=xev->xmotion.y-pressed_y+gad->y-posy;
		  gad->x=(posx+=dx);
		  gad->y=(posy+=dy);
		  checkPosition(gad,posx,posy,IS_SNAPPED(gad));
		  XMoveWindow(display,gad->win,
			      gad->x,gad->y);
		  XFlush(display);
		  if ((xev->xmotion.state&5)==5) {
		    paintInitButton(im_data,gad);
		    button_state=0;
		  }
		  usleep(50000);
		  
		} else if (pressed_b>1) {
		  int nb;
		  Gadget *all;
		  
		  assert(current_page>=0 && current_page<npages,"Page overflow");
		  
		  skip_update=((xev->xmotion.state&5)==4);
		  dx=xev->xmotion.x-pressed_x;
		  dy=xev->xmotion.y-pressed_y;
		  all=getButtons(current_page,&nb);
		  assert(all!=NULL,"No button");
		  
		  if (checkedMove(&dx,&dy,all,nb)) {
		    for(j=0;j<nb;j++) {
		      all[j].x+=dx;
		      all[j].y+=dy;
		      XMoveWindow(display,all[j].win,
				  all[j].x,
				  all[j].y);
		    }
		    XFlush(display);
		    if (!skip_update)
		      for(j=0;j<nb;j++)
			paintInitButton(im_data,all+j);
		    button_state=0;
		  }
		  usleep(160000);
		  
		}
	      }
	    } else if (xev->xmotion.state==0) { /* simple motion */
	      debugPrint(("state:%d\n",button_state));
	      if (button_state!=1) {
		paintButton(im_data,gad,FOCUSED);
		button_state=1;
		focused_gadget=gad;
	      }
	    }
	    break;
	  

	  case KeyRelease:
	    debugPrint(("\tXKey(0x%x : 0x%x)=%s\n",
			xev->xkey.keycode,
			(unsigned int)XKeycodeToKeysym(display,
						       (KeyCode) xev->xkey.keycode,0),
			XKeysymToString(XKeycodeToKeysym(display,
							 (KeyCode) xev->xkey.keycode,0))));
	    j=current_page;
	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();
	    exec_command(getKeyAction((unsigned int)
				      XKeycodeToKeysym(display,
						       (KeyCode)xev->xkey.keycode,0)),
			 gad);
	    if (j!=current_page) {
	      button_state=0;
	      focused_gadget=NULL;
	    }
	    break;
	  
	  case ClientMessage:
	    if (xev->xclient.format == 32 &&
		xev->xclient.data.l[0] == (signed)wmDeleteWindow) {
	      saveButtons();
	      exit(0);
	    }
	    break;

	  }
	} else if (IS_PERIODIC(gad)) {

	  debugPrint(("XEvent for periodic gadgets\n"));

	  switch(xev->type) {

	  case ConfigureNotify:
	    {
	      int posx=gad->x,posy=gad->y;
	      Window dummy;

	      debugPrint(("\t(x,y) = %d,%d ; widthxheight = %dx%d\n",
			  xev->xconfigure.x,xev->xconfigure.y,
			  xev->xconfigure.width,xev->xconfigure.height));
	      XTranslateCoordinates(display,gad->win,
				    XDefaultRootWindow(display),
				    0,0,&gad->x,&gad->y,&dummy);

	      assert(PERIODIC(gad).movedGadget!=NULL,"Invalid gadget");
	      assert(PERIODIC(gad).resizedGadget!=NULL,"Invalid gadget");

	      if (posx!=gad->x || posy!=gad->y)
		PERIODIC(gad).movedGadget(gad);
	      if (gad->width!=xev->xconfigure.width ||
		  gad->height!=xev->xconfigure.height) {
		gad->width=xev->xconfigure.width;
		gad->height=xev->xconfigure.height;
		PERIODIC(gad).resizedGadget(gad);
	      }
	    }
	    break;
	    
	  case ButtonPress:
	    debugPrint(("\t(x,y) = %d,%d ; button = %d ; state = %d\n",
			xev->xbutton.x,xev->xbutton.y,
			xev->xbutton.button,
			xev->xbutton.state));
	  
	    pressed_x=xev->xbutton.x;
	    pressed_y=xev->xbutton.y;
	    pressed_b=xev->xbutton.button;
	    posx=gad->x;
	    posy=gad->y;
	    button_state=-1;
	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();
	    PERIODIC(gad).pressedGadget(gad);
	    break;

	  case EnterNotify:
	    button_state=1;
	    if (SHOWS_TOOLTIP(gad)) {
	      debugPrint((GREEN"Activated tooltip"BLACK"\n"));
	      focused_gadget=gad;
	    }
	    PERIODIC(gad).enteredGadget(gad);
	    break;

	  case ButtonRelease:
	    debug( if (xev->type==ButtonRelease)
		   printf("\t(x,y) = %d,%d ; button = %d ; state = %d\n",
			  xev->xbutton.x,xev->xbutton.y,
			  xev->xbutton.button,
			  xev->xbutton.state) );
	    button_state=1;
	    skip_moves=0;
	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();
	    XFlush(display);
	    if ((xev->xbutton.state&5)==0 &&
		xev->xbutton.x>=0 && xev->xbutton.x<gad->width &&
		xev->xbutton.y>=0 && xev->xbutton.y<gad->height)
	      PERIODIC(gad).releasedGadget(gad);
	    break;

	  case LeaveNotify:
	    button_state=0;
	    focused_gadget=NULL;
	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();
	    PERIODIC(gad).leftGadget(gad);
	    break;


	  case MotionNotify:
	    debugPrint(("\t(x,y) = %d,%d ; state = %d\n",
			xev->xmotion.x,xev->xmotion.y,xev->xmotion.state));

	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();

	    if (xev->xmotion.state&(256|512|1024)) {
	      skip_moves=1;
	      if (xev->xmotion.state&4) {
		int dx,dy;
		
		debugPrint(("\t+(%d,%d)\n",dx,dy));
		dx=xev->xmotion.x-pressed_x+gad->x-posx;
		dy=xev->xmotion.y-pressed_y+gad->y-posy;
		gad->x=(posx+=dx);
		gad->y=(posy+=dy);
		checkPosition(gad,posx,posy,IS_SNAPPED(gad));
		XMoveWindow(display,gad->win,
			    gad->x,gad->y);
		XFlush(display);
		if ((xev->xmotion.state&5)==5) {
		  button_state=0;
		}
		usleep(20000);
	      }
	    } else if (xev->xmotion.state==0) { /* simple motion */
	      debugPrint(("state:%d\n",button_state));
	      if (button_state!=1) {
		button_state=1;
		focused_gadget=gad;
	      }
	    }
	    break;
	    
	  case KeyRelease:
	    debugPrint(("\tXKey(0x%x : 0x%x)=%s\n",
			xev->xkey.keycode,
			(unsigned int)XKeycodeToKeysym(display,
						       (KeyCode) xev->xkey.keycode,0),
			XKeysymToString(XKeycodeToKeysym(display,
							 (KeyCode) xev->xkey.keycode,0))));
	    j=current_page;
	    unregisterPeriodicGadget(NULL);
	    hide_tooltip();
	    exec_command(getKeyAction((unsigned int)
				      XKeycodeToKeysym(display,
						       (KeyCode)xev->xkey.keycode,0)),
			 gad);
	    if (j!=current_page) {
	      button_state=0;
	      focused_gadget=NULL;
	    }
	    break;
	  
	  case ClientMessage:
	    if (xev->xclient.format == 32 &&
		xev->xclient.data.l[0] == (signed)wmDeleteWindow) {
	      saveButtons();
	      exit(0);
	    }
	    break;

	  }
	}
      }
    }

    XFlush(display);
    if (!XPending(display)) {
      int nfd=xfd;
      if (focused_gadget &&
	  tooltip_win==0 &&
	  (button_state==1 || button_state==-1)) {
	tooltip_time=time+tooltip_period;
	trace( registerPeriodicGadget(NULL) );
      }
      FD_ZERO(&fdset);
      FD_SET(xfd,&fdset);
      trace( delay=get_delay(time,&tval,focused_gadget,&nfd,&fdset) );
      debug(if (!delay) printf("select: wait null\n");
	    else printf("select: wait %ld sec, %ld usec\n",
			delay->tv_sec,delay->tv_usec));
      if (select(nfd+1,&fdset,NULL,NULL,delay)>0) {
	/* on which file descriptor? */
	for(i=0;i<n_jobs;i++)
	  if (jobs[i].gad && PERIODIC(jobs[i].gad).calledGadget(jobs[i].gad,&fdset))
	    jobs[i].time=0.0; /* will be called soon! */
      }
    }

  }
}


