/**
 * Debugging module
 */

#include "debug.h"
#include <signal.h>
#include <string.h>

/* private functions to check for memory
 * consistency and allocation
 */

#if DEBUG

int print_active=1;

#define panicAssert(cond_expr,explanation) \
do {\
  if (!(cond_expr)) {\
    fprintf(stderr,\
	    RED"** Debug panic!!! Violated assertion(%s): "POS_STR" **\n"\
            "** %s"BLACK"\n",\
	    __STRING(cond_expr),POS_ARGS,explanation?explanation:"");\
    printStackTrace();\
    exit(1);\
  }\
} while(0)

#define stackTraceAssert(cond_expr,explanation) \
do {\
  if (!(cond_expr)) {\
    fprintf(stderr,\
	    RED"** Stack Trace panic !!! Violated assertion(%s): "POS_STR" **\n"\
            "** %s"BLACK"\n",\
	    __STRING(cond_expr),POS_ARGS,explanation?explanation:"");\
    exit(1);\
  }\
} while(0)


typedef struct {
  char *file,*function;
  int line;
  char *call,*real_function,*real_file,*args;
} FilePos;


typedef struct {
  void *p;
  int size;
  int group;
  int ignore;
  FilePos *filePos;
  int n_filePos;
} Block;

static Block *blocks=NULL;
static int n_blocks=0;
static int alloc_memory=0;
static int current_group=0;

#define IGNORE_ONCE 1
#define IGNORE_ALWAYS 2

static FilePos *call_stack=NULL;
static int stack_depth=0;

static int mem_id=0;

/* top level push/pop: for functions that cannot
 * be pushed/popped
 */
static char *top_file=NULL;
static char *top_func=NULL;
static int top_line=0;

#define TOP_PUSH() \
do {\
  top_file=in_file;\
  top_func=in_function;\
  top_line=at_line;\
  assert(at_line>0,"Invalid line number");\
} while(0)

#define TOP_PUSH_HERE() \
do {\
  top_file=__FILE__;\
  top_func=__FUNCTION__;\
  top_line=__LINE__;\
} while(0)

#define TOP_POP() \
do {\
  top_file=NULL;\
  top_func=NULL;\
  top_line=0;\
} while(0)
 

/* former SEG FAULT handler */
static void (*sigSegHandler)(int);


static void printTheStack(int depth,FilePos *stack) {
  int i;
  if (depth>0 && stack)
    for(i=depth-1;i>=0;i--)
      fprintf(stderr,
	      YELLOW "\t%s: %s(%s) called from %s { %s }, file %s, line %d"BLACK"\n",
	      stack[i].real_file?stack[i].real_file:"",
	      stack[i].real_function?stack[i].real_function:"",
	      stack[i].args?stack[i].args:"",
	      stack[i].function,
	      stack[i].call,
	      stack[i].file,
	      stack[i].line);
}


#define FREE_POS(field) if (b->filePos[i].##field) free(b->filePos[i].##field)

/**
 * clean position in a block
 */
static void clean_pos(Block *b) {
  pushMe();
  if (b) {
    int i;
    assert((b->n_filePos==0)==(b->filePos==NULL),"Inconsistent data");
    assert(b->n_filePos>=0,"Corrupted data");
    for(i=0;i<b->n_filePos;i++)
      FREE_POS(args);
    if (b->filePos) free(b->filePos);
    b->filePos=NULL;
    b->n_filePos=0;
  }
}


#define COPY_POS(field) \
 if (call_stack[i].##field) \
  b->filePos[i].##field=strdup(call_stack[i].##field); \
 else \
  b->filePos[i].##field=NULL

#define SET_POS(field) \
 b->filePos[i].##field=call_stack[i].##field

/**
 * store position in a block
 */
static void store_pos(Block *b) {
  if (b && stack_depth>0 && call_stack) {
    int i;
    if (b->filePos) clean_pos(b);
    b->filePos=(FilePos*)malloc(stack_depth*sizeof(FilePos));
    if (b->filePos) {
      b->n_filePos=stack_depth;
      for(i=0;i<stack_depth;i++) {
	SET_POS(file);
	SET_POS(function);
	SET_POS(line);
	SET_POS(call);
	SET_POS(real_function);
	SET_POS(real_file);
	COPY_POS(args);
      }
    }
  }
}


/**
 * sanity check of the block structure
 */
static void checkBlocks() {
  int i,j,checkmem=0;

  pushMe();
  assert(alloc_memory>=0,"Corrupted memory counter");
  assert(n_blocks>=0,"Corrupted block counter");
  /* n_blocks==0 => blocks==NULL */
  assert(blocks!=NULL || n_blocks==0,"Internal error: null => no block");
  /* alloc_memory==0 => n_blocks==0 */
  assert(alloc_memory!=0 || n_blocks==0,"Internal error: no memory => no block");
  
  /* check for pointer uniqueness */
  for(i=0;i<n_blocks;i++) {
    assert(blocks[i].size>0,"Invalid size <=0");
    assert(blocks[i].p!=NULL,"Invalid null pointer");
    checkmem+=blocks[i].size;
    for(j=i+1;j<n_blocks;j++)
      assert(blocks[i].p!=blocks[j].p,"Duplicate block pointer");
  }

  assert(checkmem==alloc_memory,"Inconsistent memory count");
}


/**
 * computes a sort of quick checksum
 */
static int memory_id(void *p,size_t size) {
  int i,res;
  char *ptr;

  pushArgs(32,(BUF,"%s,%d",PTR2STR(p),size));
  assert(p!=NULL || size==0,"Internal error: null => size==0");
  assert(size>=0,"Bad memory size <0");

  res=0;
  ptr=p;
  for(i=0;i<size;i++)
    res+=37*ptr[i];

  return res;
}


/**
 * return the block corresponding to
 * the memory pointer
 */
static Block* getBlock(void *p) {
  int i,found=0;
  Block *b=NULL;

  pushArgs(8,(BUF,"%s",PTR2STR(p)));
  assert(p!=NULL,"Null memory block pointer");
  assert(n_blocks>=0,"Corrupted block counter");
  assert(blocks!=NULL || n_blocks==0,"Internal error: null pointer => no block");

  if (p) {
    for(i=0;i<n_blocks;i++)
      if (blocks[i].p==p) {
	if (found==0)
	  b=&blocks[i];
	found++;
      }
    
    /* found!=0 */
    assert(found!=0,"Memory block not found, probably bad pointer");
    /* found!=0 => found==1 */
    assert(found==0 || found==1,"Block inconsistency: several blocks matched!");
  }

  return b;
}


/**
 * return size of a given block
 */
static int getBlockSize(void *p) {
  Block *b;
  int size=0;

  pushArgs(8,(BUF,"%s",PTR2STR(p)));

  trace(b=getBlock(p));
  if (b!=NULL)
    trace(size=b->size);

  return size;
}


/**
 * remove a block
 */
static int removeBlock(void *p) {
  int i,found=0;
  int group=0;

  pushArgs(8,(BUF,"%s",PTR2STR(p)));
  assert(p!=NULL,"Null memory block pointer");
  assert(n_blocks>=0,"Corrupted block counter");
  assert(blocks!=NULL || n_blocks==0,"Internal error: null pointer => no block");

  for(i=0;i<n_blocks;i++)
    if (blocks[i].p==p) {
      if (found==0) {
	int j,k;
	trace(clean_pos(blocks+i));
	alloc_memory-=blocks[i].size;
	group=blocks[i].group;
	for(j=i;j<n_blocks-1;j++) /* move */
	  blocks[j]=blocks[j+1];
	n_blocks--;
	trace(j=memory_id(blocks,n_blocks*sizeof(Block)));
	trace(blocks=(Block*)realloc(blocks,n_blocks*sizeof(Block)));
	panicAssert(blocks!=NULL || n_blocks==0,
		    "Internal error: null pointer => no block");
	trace(k=memory_id(blocks,n_blocks*sizeof(Block)));
	assert(j==k,"Corrupted memory content");
      }
      found++;
    }
  
  assert(found==1,
	 found<1?
	 "Memory block not found, probably bad pointer":
	 "Duplicate pointers found, corrupted table");
  return group;
}


/**
 * add a block
 */
static Block* addBlock(void *p,size_t size) {
  int i,j;

  pushArgs(32,(BUF,"%s,%d",PTR2STR(p),size));
  assert(p!=NULL,"Null memory block pointer");
  assert(n_blocks>=0,"Corrupted block counter");
  assert(blocks!=NULL || n_blocks==0,"Internal error: null pointer => no block");
  
  trace(i=memory_id(blocks,n_blocks*sizeof(Block)));
  trace(blocks=(Block*)realloc(blocks,(n_blocks+1)*sizeof(Block)));
  panicAssert(blocks!=NULL,"Internal error: blocks not reallocated");
  trace(j=memory_id(blocks,n_blocks*sizeof(Block)));
  assert(i==j,"Corrupted memory content");

  blocks[n_blocks].p=p;
  blocks[n_blocks].size=size;
  blocks[n_blocks].group=current_group;
  blocks[n_blocks].ignore=0;
  blocks[n_blocks].filePos=NULL;
  blocks[n_blocks].n_filePos=0;
  alloc_memory+=size;
  n_blocks++;
  return &blocks[n_blocks-1];
}


/***********
 * private *
 ***********/

void intern_pushCall(char *string_call,POS_VARS) {
  assert(call_stack!=NULL || stack_depth==0,
	 "Internal error: pointer null => empty call stack");
  assert(in_file!=NULL,"Internal error: null argument");
  assert(in_function!=NULL,"Internal error: null argument");
  assert(at_line>0,"Internal error: bad line number");
  assert(string_call!=NULL,"Internal error: null argument");
  
  TOP_PUSH();
  stack_depth++;
  call_stack=(FilePos*)realloc(call_stack,stack_depth*sizeof(FilePos));
  panicAssert(call_stack!=NULL || stack_depth==0,
	      "Internal error: null pointer => empty call stack");
  call_stack[stack_depth-1].file=in_file;
  call_stack[stack_depth-1].function=in_function;
  call_stack[stack_depth-1].line=at_line;
  /* the following cheats and try to give the macro name, not the function name */
  call_stack[stack_depth-1].call=
    ( strncmp(string_call,"intern_",7) ? string_call : string_call+7);
  call_stack[stack_depth-1].real_function=NULL;
  call_stack[stack_depth-1].real_file=NULL;
  call_stack[stack_depth-1].args=NULL;
  TOP_POP();
}


void intern_popCall(POS_VARS) {
  assert(call_stack!=NULL || stack_depth==0,
	 "Internal error: null pointer => empty call stack");
  assert(stack_depth>0,"Bad call error: empty call stack");
  
  TOP_PUSH();
  if (stack_depth>0) {
    stack_depth--;
    if (call_stack[stack_depth].args)
      free(call_stack[stack_depth].args);
    call_stack=(FilePos*)realloc(call_stack,stack_depth*sizeof(FilePos));
    panicAssert(call_stack!=NULL || stack_depth==0,
		"Internal error: null pointer => empty call stack [realloc failed]");
  }
  TOP_POP();
}


void intern_pushArgs(char *args,char *in_file,char *in_function) {
  /* args can be NULL */
  assert(stack_depth>0 && call_stack!=NULL,
	 stack_depth==0 ?
	 "Bad pushArgs usage: empty call stack, use trace(<CALL>) or pushCall(<CALL_STRING>) before":
	 "Corrupted stack depth <0");
  assert(call_stack!=NULL || stack_depth==0,
	 "Internal error: null pointer => empty call stack");

  TOP_PUSH_HERE();
  if (stack_depth>0) {
    if (call_stack[stack_depth-1].args)
      free(call_stack[stack_depth-1].args);
    call_stack[stack_depth-1].real_function=in_function;
    call_stack[stack_depth-1].real_file=in_file;
    call_stack[stack_depth-1].args=(args==NULL?NULL:strdup(args));
  }
  TOP_POP();
}


static void debugSigSeg(int sig) {
  fprintf(stderr,RED"Segmentation fault:\n");
  if (sigSegHandler)
    signal(SIGSEGV,sigSegHandler);
  else
    signal(SIGSEGV,exit);
  printStackTrace();

  if (sigSegHandler)
    sigSegHandler(sig);
  else
    exit(1);
}

static int computeMemID() {
  int i,res=0;

  pushMe();
  trace(checkBlocks());

  for(i=0;i<n_blocks;i++)
    if (!blocks[i].ignore)
      trace(res=res*37+memory_id(blocks[i].p,blocks[i].size));

  return res;
}


/***********
 *** API ***
 ***********/


void intern_checkAlloc(void *ptr) {
  pushArgs(8,(BUF,"%s",PTR2STR(ptr)));
  trace(checkBlocks());
  trace(getBlock(ptr));
}


void intern_checkPtr(void *p) {
  int i,found=0;

  pushArgs(8,(BUF,"%s",PTR2STR(p)));
  trace(checkBlocks());
  assert(p!=NULL,"Bad pointer: NULL");

  if (p) {
    unsigned int m=(unsigned int)p;
    for(i=0;i<n_blocks;i++) {
      unsigned int start=(unsigned int)blocks[i].p;
      unsigned int end=start+blocks[i].size;
      if (m>=start && m<end)
	found++;
    }
    
    assert(found!=0,"Bad pointer: outside allocated memory");
    /* found!=0 => found==1 */
    assert(found==0 || found==1,"Block inconsistency: overlapped memory blocks");
  }
}


void intern_addMemoryBlock(void *ptr,size_t size) {
  pushArgs(256,(BUF,"%s,%d",PTR2STR(ptr),size));
  assert(size>=0,"Bad size value");
  assert(ptr==NULL || size>0,"Inconsistent arguments");
  if (ptr) {
    Block *block;
    trace(block=addBlock(ptr,size));
    store_pos(block);
  }
}


void intern_addStringBlock(char *ptr) {
  pushArgs(STRSIZE(ptr)+8,(BUF,"char*=%s",ptr));
  if (ptr)
    addMemoryBlock(ptr,strlen(ptr)+1);
}


void printStackTrace() {
  /* no assert here, otherwise recursive loop */
  stackTraceAssert(call_stack!=NULL || stack_depth==0,
		   "Internal error: null pointer => empty call stack [realloc failed]");
  
  assert((top_file==NULL && top_func==NULL && top_line==0) ||
	 (top_file!=NULL && top_func!=NULL && top_line!=0),
	 "Internal error: corrupted top level call information");

  if (top_file!=NULL && top_func!=NULL)
    fprintf(stderr,
	    RED "\tfrom %s(), file %s, line %d"BLACK"\n",
	    top_func,top_file,top_line);

  printTheStack(stack_depth,call_stack);
}


void *alloc(size_t size) {
  void *p;

  pushArgs(32,(BUF,"%d",size));
  assert(size>=0,"Invalid size <0");
  trace(checkBlocks());

  trace(p=malloc(size));

  if (p!=NULL) {
    Block *block;
    trace(block=addBlock(p,size));
    store_pos(block);
  }
  trace(checkBlocks());

  return p;
}


void *nalloc(size_t nelem,size_t size) {
  void *p;

  pushArgs(64,(BUF,"%d,%d",nelem,size));
  assert(size>=0,"Invalid size <0");
  assert(nelem>=0,"Invalid number of elements <0");
  trace(checkBlocks());

  trace(p=calloc(nelem,size));

  if (p!=NULL) {
    Block *block;
    trace(block=addBlock(p,size));
    store_pos(block);
  }
  trace(checkBlocks());

  return p;
}


void *ralloc(void *ptr,size_t size) {
  void *p;
  int check=0,check2=0;
  int smaller=0,group=current_group;
  Block *b;

  pushArgs(40,(BUF,"%s,%d",PTR2STR(ptr),size));
  assert(size>=0,"Invalid size <0");
  trace(checkBlocks());
  
  if (ptr!=NULL) {
    int old_group;
    trace(smaller=getBlockSize(ptr));
    if (smaller>size)
      smaller=size;
    trace(check=memory_id(ptr,smaller));
    trace(old_group=removeBlock(ptr));
    assert(old_group==group,
	   "Bad memory group management: reallocated group has a different group than the current one");
    group=old_group; /* try to recover if pb */
  }

  trace(p=realloc(ptr,size));

  trace(check2=memory_id(p,smaller));
  assert(check==check2,"Corrupted memory content");
  trace(b=addBlock(p,size));
  store_pos(b);
  assert(b!=NULL,"Internal error: added block is null");
  trace(b->group=group);
  trace(checkBlocks());

  return p;
}


void intern_disalloc(void *ptr) {
  pushArgs(8,(BUF,"%s",PTR2STR(ptr)));
  assert(ptr!=NULL,"Free a null pointer");
  trace(checkBlocks());
  if (ptr!=NULL)
    trace(removeBlock(ptr));
  trace(free(ptr));
  trace(checkBlocks());
}


void intern_checkFreed() {
  int i;

  pushMe();
  trace(checkBlocks());
  assert(alloc_memory==0,"Memory leak: memory has not been desallocated");
  if (alloc_memory!=0)
    trace(fprintf(stderr,RED"Memory left: %d bytes"BLACK"\n",alloc_memory));

  assert((blocks==NULL)==(n_blocks==0),"Inconsistent data");
  assert(n_blocks>=0,"Corrupted data");
  for(i=0;i<n_blocks;i++) {
    fprintf(stderr,
	    RED"%d bytes were allocated at\n",
	    blocks[i].size);
    printTheStack(blocks[i].n_filePos,blocks[i].filePos);
  }
}


void intern_printMemory() {
  pushMe();
  trace(checkBlocks());
  trace(printf("Allocated memory: %d bytes\n",alloc_memory));
  trace(fflush(stdout));
}


void intern_setMemoryGroup(int id) {
  current_group=id;
}


void intern_checkGroupIDFreed(int id) {
  int i,sum=0;

  pushArgs(32,(BUF,"%d",id));
  trace(checkBlocks());
  for(i=0;i<n_blocks;i++) {
    assert(blocks[i].group!=id,"Memory leak: a memory group is left");
    if (blocks[i].group==id)
      sum+=blocks[i].size;
  }
  assert(sum==0,"Memory leak: memory group has not been desallocated");
  if (sum!=0)
    trace(fprintf(stderr,RED"Memory left for group %d: %d bytes"BLACK"\n",id,sum));
}


void intern_checkGroupFreed() {
  pushMe();
  trace(checkGroupIDFreed(current_group));
}


void initDebug() {
  assert(sigSegHandler==NULL,"Bad call: the SIGSEGV handler was already saved");
  sigSegHandler=signal(SIGSEGV,debugSigSeg);
}


void intern_updateMem() {
  pushMe();
  trace(mem_id=computeMemID());
}


void intern_checkMem() {
  int new_id;
  pushMe();
  trace(new_id=computeMemID());
  assert(new_id==mem_id,"Memory has changed! Corrupted memory or updateMem() was not called.");
}


void intern_ignoreMem(void *p) {
  Block *b;
  
  pushArgs(32,(BUF,"%s",PTR2STR(p)));

  trace(b=getBlock(p));
  if (b)
    trace(b->ignore=1);

  updateMem();
}


void intern_dontIgnoreMem(void *p) {
  Block *b;

  pushArgs(32,(BUF,"%s",PTR2STR(p)));

  trace(b=getBlock(p));
  if (b)
    trace(b->ignore=0);

  updateMem();
}


#endif


