/* debug.h
 * debugging facilities
 */

#ifndef DEBUG_H

#include <stdio.h>
#include <stdlib.h>

#ifndef DEBUG
#define DEBUG 1
#endif

#define BLACK   "\033[00m"
#define RED     "\033[0;30;41m"
#define GREEN   "\033[0;30;42m"
#define YELLOW  "\033[0;30;43m"
#define BLUE    "\033[0;30;44m"
#define MAGENTA "\033[0;30;45m"
#define CYAN    "\033[0;30;46m"
#define WHITE   "\033[0;30;47m"

#if DEBUG

extern int print_active;

/* position macros */
#define POS_ARGS     __FILE__,__FUNCTION__,__LINE__
#define POS_VARS     char *in_file,char *in_function,int at_line
#define POS_STR      "file %s, function %s(), line %d"

/* ptr to string check */
#define PTR2STR(p) ((p)!=NULL?"<PTR>":"NULL")
#define STRSIZE(p) ((p)?strlen(p):4)

/* print location of the line */
#define printHere() \
do {\
  printf(POS_STR": ",POS_ARGS);\
  fflush(stdout);\
} while(0)


#define printEnable() print_active=1
#define printDisable() print_active=0

/* normal debug print */
#define debugPrint(s) \
do {\
  if (print_active) {\
    printf s;\
    fflush(stdout);\
  }\
} while(0)

/* debug instruction */
#define debug(def) def


/* push the current call position */
#define pushCall(string_call) intern_pushCall(string_call,POS_ARGS)


/* pop the call position */
#define popCall()  intern_popCall(POS_ARGS)


/* trace a call */
#define trace(call) do { pushCall(__STRING(call)); call; popCall(); } while(0)


/* remember args optionally
 * use: pushArgs(256,(BUF,"%d,%d",i,j)); */
#define pushArgs(buf_size,formatted_args)\
do {\
  char BUF[buf_size];\
  sprintf formatted_args;\
  intern_pushArgs(BUF,__FILE__,__FUNCTION__);\
} while(0)


/* remember function name */
#define pushMe() intern_pushArgs(NULL,__FILE__,__FUNCTION__)


/* condition verification */
#define assert(cond_expr,explanation) \
do {\
  if (!(cond_expr)) {\
    fprintf(stderr,\
	    RED"** Violated assertion(%s): "POS_STR" **\n"\
	    BLACK"\t"CYAN"%s"BLACK"\n",\
	    __STRING(cond_expr),POS_ARGS,explanation?explanation:"");\
    printStackTrace();\
  }\
} while(0)


/* normal memory management, instead of
 * malloc, realloc, free and calloc
 * exactly the same behaviour
 */
#define disalloc(ptr)    trace(intern_disalloc(ptr))
/* use these with trace( blabla=alloc(a_size) );
 * followed by updateMem() */
void *alloc(size_t size);
void *ralloc(void *ptr,size_t size);
void *nalloc(size_t nelem,size_t size);

/* register block of memory, allocated by
 * another call, e.g. strdup */
#define addMemoryBlock(ptr,size) trace(intern_addMemoryBlock(ptr,size))
#define addStringBlock(ptr) trace(intern_addStringBlock(ptr))

/* total memory check */
#define checkFreed()     trace(intern_checkFreed())
#define printMemory()    trace(intern_printMemory())

/* check if the pointer is valid: points in an allocated memory zone */
#define checkPtr(p)      trace(intern_checkPtr(p))
/* check if the pointer is valid: valid allocated pointer */
#define checkAlloc(p)    trace(intern_checkAlloc(p))

/* memory group check */
#define setMemoryGroup(id)    trace(intern_setMemoryGroup(id))
#define checkGroupIDFreed(id) trace(intern_checkGroupIDFreed(id))
#define checkGroupFreed()     trace(intern_checkGroupFreed())


void printStackTrace();
void initDebug(); /* should be called at the beginning */

/* update memory signature */
#define updateMem() trace(intern_updateMem())
/* check memory signature */
#define checkMem()  trace(intern_checkMem())
/* ignore memory check */
#define ignoreMem(p) trace(intern_ignoreMem(p))
/* reset */
#define dontIgnoreMem(p) trace(intern_dontIgnoreMem(p))

/* internal functions ; do not use them */
void intern_addMemoryBlock(void *ptr,size_t size);
void intern_addStringBlock(char *ptr);
void intern_pushCall(char *string_call,POS_VARS);
void intern_popCall(POS_VARS);
void intern_pushArgs(char *args,char *in_file,char *in_function);
void intern_disalloc(void *ptr);
void intern_checkFreed();
void intern_printMemory();
void intern_setMemoryGroup(int id);
void intern_checkGroupIDFreed(int id);
void intern_checkGroupFreed();
void intern_updateMem();
void intern_checkMem();
void intern_ignoreMem(void *p);
void intern_dontIgnoreMem(void *p);
void intern_checkPtr(void *p);
void intern_checkAlloc(void *p);

#else


#define printHere()
#define debugPrint(s)
#define debug(def)

#define pushMe()
#define pushCall()
#define popCall()
#define pushArgs(a,f)
#define trace(call) call
#define assert(c,s)

#define alloc(size)      malloc(size)
#define ralloc(ptr,size) realloc(ptr,size)
#define disalloc(ptr)    free(ptr)
#define addMemoryBlock(p,s)
#define addStringBlock(p)

#define checkPtr(p)
#define checkAlloc(p);

#define checkFreed()
#define printMemory()

#define setMemoryGroup(id)
#define checkGroupIDFreed(id)
#define checkGroupFreed()

#define printStackTrace()
#define initDebug()

#define updateMem()
#define checkMem()
#define ignoreMem(p)
#define dontIgnoreMem(p)

#define printEnable()
#define printDisable()

#endif /* DEBUG */

#define resetGroup() setMemoryGroup(0)
#define debugOK() debugPrint(("OK\n"))


#endif /* DEBUG_H */









