/* font.c
 * part of FancyLauncher
 * initialize font info from
 * a given font name and draw
 * text on ImlibImage
 *
 * license: GPL
 * author: Alexandre David
 * e-mail: adavid@docs.uu.se
 * http://www.docs.uu.se/~adavid
 */

#include "debug.h"
#include "font.h"
#include "pixmaps.h"
#include "utils.h"
#include <X11/Xlib.h>


FontInfo **loaded_fonts=NULL;
int n_fonts=0;


void free_fonts() {
  int i;
  assert(loaded_fonts!=NULL || n_fonts==0,"Inconsistent data");
  for(i=0;i<n_fonts;i++) {
    if (loaded_fonts[i]) {
      FREE(loaded_fonts[i]->name);
      FREE(loaded_fonts[i]->data);
      disalloc(loaded_fonts[i]);
    }
  }
  FREE(loaded_fonts);
  loaded_fonts=NULL;
  n_fonts=0;
}


#if DEBUG
void printFont(FontInfo *font) {
  int i;
  printf("[%s] %s\n",(*font->name=='0'?"normal":"bold1st"),font->name+1);
  for(i=0;i<NB_CHARS;i++) {
    int j,count;
    unsigned int *data=font->data+i*font->size;
    unsigned int code;
    printf("                               %c\n",(char)(i+FIRST_CHAR));
    for(j=0,count=0;j<font->height;j++) {
      int col;
      for(col=0;col<font->width;col++,count++,code>>=1) {
	if ((count&31)==0)
	  code=*data++;
	if (code&1)
	  printf("*");
	else
	  printf(" ");
      }
      printf("\n");
    }
  }
}
#endif


/* load a font and converts the pixels
 * in bit information in order to draw
 * the them manually on any ImlibImage
 */
FontInfo* loadFont(char *name,int boldFirst) {
  XImage *fontImage;
  XFontStruct *xfont=NULL;
  XGCValues gc_val;
  Pixmap fontPixmap=0;
  GC gc;
  FontInfo *font;
  int size,width,height,i,ascent,line_size;
  char *bname,*data;
  unsigned int *dst;

  if (!name)
    name="6x*";
  trace(bname=(char*)safeAllocPtr(alloc(strlen(name)+2)));
  *bname='0'+(boldFirst!=0);
  trace(strcpy(bname+1,name));

  assert((n_fonts==0)==(loaded_fonts==NULL),"Inconsistent data");
  for(i=0;i<n_fonts;i++)
    if (!strcmp(bname,loaded_fonts[i]->name)) {
      disalloc(bname);
      return loaded_fonts[i];
    }
  xfont=XLoadQueryFont(display,name);

  if (xfont==NULL) /* take default */
    xfont=XLoadQueryFont(display,"6x*");
  
  if (xfont==NULL) {
    fprintf(stderr,"Could not load font, aborting!\n");
    exit(1);
  }

  width=xfont->max_bounds.width;
  ascent=xfont->max_bounds.ascent;
  height=ascent+xfont->max_bounds.descent;
  size=(width*height+31)/32;

  n_fonts++;
  trace(loaded_fonts=(FontInfo**)safeAllocPtr(ralloc(loaded_fonts,n_fonts*sizeof(FontInfo*))));
  trace(font=(FontInfo*)safeAllocPtr(alloc(sizeof(FontInfo))));
  loaded_fonts[n_fonts-1]=font;
  font->width=width;
  font->height=height;
  font->size=size;
  font->name=bname;
  trace(font->data=(unsigned int*)safeAllocPtr(alloc(size*NB_CHARS*sizeof(unsigned int))));
  
  fontPixmap=XCreatePixmap(display,im_data->x.root,
			   width*NB_CHARS,height,1);//im_data->x.depth);
  if (fontPixmap==0) {
    fprintf(stderr,"Could not create pixmap!\n");
    exit(1);
  }      
  
  gc_val.foreground=0;//BlackPixel(display,DefaultScreen(display));
  gc_val.background=0;//BlackPixel(display,DefaultScreen(display));
  gc_val.fill_style=FillSolid;
  gc_val.function=GXcopy;
  gc=XCreateGC(display,fontPixmap,
	       GCFillStyle|GCFunction|GCForeground|GCBackground,
	       &gc_val);
  if (gc==0) {
    fprintf(stderr,"Could not create graphic context\n");
    exit(1);
  }
  XFillRectangle(display,fontPixmap,gc,0,0,width*NB_CHARS,height);
  XFreeGC(display,gc);

  gc_val.font=xfont->fid;
  gc_val.foreground=~0L;//WhitePixel(display,DefaultScreen(display));
  gc=XCreateGC(display,fontPixmap,GCFont|GCForeground|GCBackground,&gc_val);
  if (gc==0) {
    fprintf(stderr,"Could not create graphic context\n");
    exit(1);
  } else {
    int x;
    unsigned char c;
    for(c=FIRST_CHAR,x=0;c!=(LAST_CHAR+1)%256/* 255+1=0 */;c++,x+=width)
      trace(XDrawString(display,fontPixmap,gc,x,ascent,&c,1));
  }
  fontImage=XGetImage(display,fontPixmap,
		      0,0,width*NB_CHARS,height,
		      AllPlanes,XYPixmap);
  if (fontImage==0) {
    fprintf(stderr,"Could not create image!\n");
    exit(1);
  }

  boldFirst&=width>2;
  line_size=fontImage->bytes_per_line;
  height*=line_size;
  data=fontImage->data;
  dst=font->data;

  for(i=0;i<NB_CHARS;i++) {
    unsigned int code=0,k=0,index=i*size,y;

    for(y=0;y<height;y+=line_size) {
      int x=i*width,maxcol=x+width;

      trace(if (boldFirst) {
	int *pix=(int*)(data+y+(((x+1)>>3)&~3));
	/* for MSBFirst: ((1<<31)>>((x+1)&31)) */
	*pix|=(*pix & (1<<((x+1)&31)))<<1;
      });

      for(;x<maxcol;x++,code>>=1) {
	//trace(if (data[y+x/8]&(1<<(x%8)))
	//for MSBFirst as on Sun machine, change:  ((1<<31)>>(x&31))
	trace(if ( (*((int*)(data+y+((x>>3)&~3))) & (1<<(x&31))) )
	      code|=(1<<31));
	k++;
	if ((k&31)==0) {
	  assert(index<size*NB_CHARS,"Overflow");
	  trace(dst[index++]=code);
	}
      }
    }
    if ((k&31)!=0) {
      assert(index<size*NB_CHARS,"Overflow");
      for(k++;(k&31)!=0;k++)
	code>>=1;
      trace(dst[index]=code);
    }
  }

  XFreeGC(display,gc);
  XFreeFont(display,xfont);
  XFreePixmap(display,fontPixmap);
  XDestroyImage(fontImage);

  debug(printFont(font));
  debugPrint(("Loaded font: %s %dx%d\n",
	      font->name,
	      font->width,
	      font->height));

  return font;
}

