/* button.c
 * core of the buttons
 * Part of FancyLauncher
 *
 * author: Alexandre David
 * e-mail: adavid@docs.uu.se
 */

#include "debug.h"
#include "utils.h"
#include "gadget.h"
#include "pixmaps.h"
#include "codes.h"
#include "Xcore.h"
#include "font.h"
#include "tooltips.h"

int npages=0,current_page=0;
ButtonPage *pages=NULL;
int npaths=0,longest_path_size=0;
char **icon_paths=NULL;

int darker=256-64,lighter=256+64;

int nkeybindings;
KeyBinding *keybindings;

InitCommand *init_commands=NULL;
int nb_init=0;

Gadget *extGadgets=NULL;
int nb_extGadgets=0;


const OptionRC buttonRC[NB_OPT]={
  { STICKY , "sticky" ,
    "The button may be sticky (1) or not (0)" },
  { BELOW , "below" ,
    "The button may be below other windows (1) or above (2) or normal (0)" },
  { SNAP , "snap" ,
    "This is to snap the position of the button" },
  { 0 , "" , 0 }, /* mapped */
  { 0, "" , 0 }, /* moving */
  { TOOLTIPS, "tooltips" ,
    "This enables the tooltips" },
  { RAISED , "raised" ,
    "This enables the raised border style" },
  { ETCHED, "etched" ,
    "This enables the etched border style" },
  { TRANSIENT, "transient" ,
    "This sets the borders transient" },
  { RESTORE, "restore" ,
    "This will restore the pixmap when the button is focused" },
  { BLEND_ICON, "blend_icon" ,
    "This is the blend factor of the icon with the background" },
  { BLEND_BACKGROUND, "blend_background" ,
    "This is the blend factor of the background+icon with the root" },
  { BLEND_FOCUS, "blend_focus" ,
    "This is the blend factor of the restored background+icon with the root" },
  { PAGE, "page" ,
    "This says in which page the button is set"},
  { 0, "" , 0 } /* nactions */
};

const char *labelnames[8]={
  "color_red_start","color_red_end",
  "color_green_start","color_green_end",
  "color_blue_start","color_blue_end",
  "posx","posy"};


Gadget* getButtons(const int p,int *n) {
  pushArgs(64,(BUF,"%d,%s",p,PTR2STR(n)));

  if (p<0 || p>=npages)
    return NULL;

  assert((npages==0)==(pages==NULL),"Inconsistent data");
  assert(n!=NULL,"Page number pointer is null");

  trace( *n=pages[p].nbuttons );
  return pages[p].buttons;
}


char* getAction(const Gadget* button,const int i) {  /* action or NULL */
  pushArgs(64,(BUF,"%s,%d",PTR2STR(button),i));
  assert(button!=NULL,"Null button pointer");
  assert(IS_BUTTON(button),"Bad type");
  assert((NACTIONS(button)==0)==(BUT(button).action==NULL),"Inconsistent data");

  if (i<0 || i>=NACTIONS(button))
    return NULL;

  return GETCHARS(BUT(button).action[2*i]);
}


char* getKeyAction(unsigned int code) {
  int i;

  pushArgs(32,(BUF,"%d",code));
  debugPrint(("getKeyAction(0x%x)\n",code));

  for(i=0;i<nkeybindings;i++)
    if (code==KEYCODE(keybindings[i]))
      return KEYCOMMAND(keybindings[i]);

  return NULL;
}


char* getTooltip(const Gadget* button,const int i) { /* tooltip or NULL */
  pushArgs(64,(BUF,"%s,%d",PTR2STR(button),i));
  assert(button!=NULL,"Null button pointer");
  assert(IS_BUTTON(button),"Bad type");
  assert((NACTIONS(button)==0)==(BUT(button).action==NULL),"Inconsistent data");

  if (i<0 || i>=NACTIONS(button))
    return NULL;

  return GETCHARS(BUT(button).action[2*i+1]);
}


ImlibImage* getPixmapBg(ImlibData *id,const Gadget* button) {
  ImlibImage *im;

  assert(button!=NULL,"Null button pointer");
  assert(id!=NULL,"Null image data pointer");

  trace( im=
    HAS_PIXMAP_BG(button) ?
    getImage(id,CHARS(GET_PIXMAP(button->bg)),button->width,button->height) :
    NULL );

  return im;
}


static ImlibImage* getButtonPixmap(ImlibData *id,const Gadget* button) {
  ImlibImage *im;

  assert(button!=NULL,"Null button pointer");
  assert(button==NULL || IS_BUTTON(button),"Gadget type is not button");
  assert(id!=NULL,"Null image data pointer");

  trace( im=
    BUT(button).icon ?
    getImage(id,CHARS(BUT(button).icon),button->width,button->height) :
    NULL );

  return im;
}


/* initialize the pixmaps : force paintInitButton
 */
void paintReInitButton(ImlibData *id,Gadget *button) {
  pushArgs(32,(BUF,"id=%s,button=%s",PTR2STR(id),PTR2STR(button)));
  assert(button!=NULL,"Null button pointer");
  assert(id!=NULL,"Null image data pointer");
  assert(button==NULL || IS_BUTTON(button),"Gadget type is not button");

  if (BUT(button).im_orig) {
    trace( Imlib_kill_image(id,BUT(button).im_orig) );
    BUT(button).im_orig=NULL;
  }
  trace( paintInitButton(id,button) );
}


#define assertSize(im) \
 assert(im==NULL || (im->rgb_width==button->width && im->rgb_height==button->height),\
        "Image dimensions are different")


/* initialize the pixmaps : paint a button
 * FIX ME : there seems to be a memory leak from this function
 */
void paintInitButton(ImlibData *id,Gadget *button) {
  pushArgs(32,(BUF,"id=%s,button=%s",PTR2STR(id),PTR2STR(button)));
  assert(button!=NULL,"Null button pointer");
  assert(id!=NULL,"Null image data pointer");
  assert(button==NULL || IS_BUTTON(button),"Gadget type is not button");
  debugPrint(("paintInitButton(%s)\n",CHARS(BUT(button).icon)));

  if (BUT(button).im_orig==NULL) { /* then make it */
    ImlibImage *icon,*bg;

    trace( icon=getButtonPixmap(id,button) );
    trace( bg=getPixmapBg(id,button) );
    assertSize(icon);
    assertSize(bg);
    
    if (bg==NULL &&          /* transparent or failed pixmap */
	(icon==NULL ||       /* because failed pixmap and moreover no icon */
	 HAS_SOLID_COLOR_BG(button))) { /* because color */
      trace( bg=getSolidImage(button->width,button->height,id,
			      HAS_SOLID_COLOR_BG(button)
			      ? GET_COLOR(button->bg)
			      : 0x808080) );
    }

    /* free masks */
    if (BUT(button).bg_mask) {
      trace( Imlib_free_pixmap(id,BUT(button).bg_mask) );
      BUT(button).bg_mask=0;
    }
    if (BUT(button).icon_mask) {
      trace( Imlib_free_pixmap(id,BUT(button).icon_mask) );
      BUT(button).icon_mask=0;
    }

    /* it would be better as this point to compute
     * a union shape for the icon and the background
     * but XShapeCombineShape does not seem to work
     * if anyone knows how to use it, drop me a note.
     */

    if (IS_BORDER_TRANSIENT(button) || !HAS_BORDER(button)) {
      if (bg) {
	Imlib_render(id,bg,button->width,button->height);
	BUT(button).bg_mask=Imlib_copy_mask(id,bg);
      }
      if (icon) {
	Imlib_render(id,icon,button->width,button->height);
	BUT(button).icon_mask=Imlib_copy_mask(id,icon);
      }
    }

    if (bg && icon) {
      trace( copyImlibImage(id,BUT(button).im_orig=bg,icon,BLEND_ICON_BG(button)) );
      trace( Imlib_kill_image(id,icon) );
      Imlib_changed_image(id,bg);
    } else if (bg==NULL)
      BUT(button).im_orig=icon;
    else if (icon==NULL)
      BUT(button).im_orig=bg;
  }

  KILL_IM(id,button,im_normal);

  trace( BUT(button).im_normal=snapshotImage(id,button->x,button->y,
						button->width,button->height,
						button->win) );
  if (BUT(button).im_normal==NULL)
    return; /* abort ; should never happen */

  if (BUT(button).im_orig)
    trace( copyImlibImage(id,
			  BUT(button).im_normal,
			  BUT(button).im_orig,
			  BLEND_BG_ROOT(button)) );

  if (!IS_BORDER_TRANSIENT(button)) {
    if (IS_BORDER_RAISED(button))
      trace( paintBorder(BUT(button).im_normal,
			 button->width,button->height,
			 RAISE_BORDER) );
    else if (IS_BORDER_ETCHED(button))
      trace( paintBorder(BUT(button).im_normal,
			 button->width,button->height,
			 ETCH_BORDER) );
  }

  trace( paintButton(id,button,NORMAL) );
}


/* paint a button in a state
 */
void paintButton(ImlibData *id,Gadget *button,ButtonState state) {
  ImlibImage *im=NULL;
  int killImage,force_rect_shape,op;

  pushArgs(32,(BUF,"%s,%s,%d",PTR2STR(id),PTR2STR(button),state));
  assert(id!=NULL,"Null image data pointer");
  assert(button!=NULL,"Null button pointer");
  assert(button==NULL || IS_BUTTON(button),"Gadget type is not button");

  debugPrint(("paintButton(%s,%s)\n",
	      BUT(button).icon ?
	      CHARS(BUT(button).icon) :
	      "null",
	      ( state==NORMAL ?
		"NORMAL" :
		(state==FOCUSED ?
		 "FOCUSED" : "PRESSED" ))));

  if (state==NORMAL) {
  
    if (BUT(button).im_snap) {
      trace( Imlib_kill_image(id,BUT(button).im_snap) );
      BUT(button).im_snap=NULL;
    }

    im=BUT(button).im_normal;
    killImage=0;
    force_rect_shape=0;
    
  } else {

    if (BUT(button).im_snap==NULL)
      trace( BUT(button).im_snap=snapshotImage(id,button->x,button->y,
						  button->width,button->height,0) );
    if (BUT(button).im_snap)
      trace( im=Imlib_clone_image(id,BUT(button).im_snap) );

    if (im!=NULL) {
      killImage=1;

      if (FOCUS_RESTORES_ICON(button) && BUT(button).im_orig)
	trace( copyImlibImage(id,im,BUT(button).im_orig,BLEND_WHEN_FOCUSED(button)) );

      trace( applyEffect(im,button->width,button->height,
			 state==FOCUSED ? lighter : darker,
			 HAS_BORDER(button) ? 2 : 0) );

      if (IS_BORDER_RAISED(button)) {

	if (state==FOCUSED)
	  trace( paintBorder(im,button->width,button->height,RAISE_BORDER) );
	else
	  trace( paintBorder(im,button->width,button->height,LOWER_BORDER) );

      } else if (IS_BORDER_ETCHED(button))
	trace( paintBorder(im,button->width,button->height,ETCH_BORDER) );
      
      force_rect_shape=(IS_BORDER_TRANSIENT(button) | HAS_BORDER(button));
    } else {
      killImage=0;
      force_rect_shape=1;
    }
  }

  /* set background */
  if (im) {
    Pixmap mask;
    /* do not use Imlib_apply_image because it is going to
     * reshape, which we want to do only once
     */
    trace( Imlib_render(id,im,button->width,button->height) );
    trace( mask=Imlib_move_image(id,im) );
    XSetWindowBackgroundPixmap(display,button->win,mask);
    trace( if (mask) Imlib_free_pixmap(id,mask) );
    trace( if (killImage) Imlib_kill_image(id,im) );
  }

  /* reshape */
  op=ShapeSet;
  if (!force_rect_shape) {
    if (BUT(button).icon_mask) {
      trace( XShapeCombineMask(display,button->win,ShapeBounding,
			       0,0,BUT(button).icon_mask,ShapeSet) );
      op=ShapeUnion;
    }

    if (BUT(button).bg_mask) {
      trace( XShapeCombineMask(display,button->win,ShapeBounding,
			       0,0,BUT(button).bg_mask,op) );
      op=ShapeUnion;
    }
  }

  if (force_rect_shape || op==ShapeSet) {
    XRectangle rect={0,0,button->width,button->height};
    trace( XShapeCombineRectangles(display,button->win,ShapeBounding,
				   0,0,&rect,1,ShapeSet,0) );
  }

  XClearWindow(display,button->win);
  XSync(display,False);
}


/* free the content of a button, not the button itself
 */
void freeButtonContent(Gadget *button) {
  int i,n;

  pushArgs(16,(BUF,"%s",PTR2STR(button)));
  assert(button!=NULL,"Null button pointer");
  assert(IS_BUTTON(button),"Bad type: button expected");
  debugPrint(("freeButtonContent()\n"));

  if (HAS_PIXMAP_BG(button))
    FREE_REF_STRING(GET_PIXMAP(button->bg));

  FREE_REF_STRING(BUT(button).icon);

  /* naction == 0 <=> action == NULL */
  assert((NACTIONS(button)==0) == (BUT(button).action==NULL),
	 "Inconsistent data");
  
  n=2*NACTIONS(button);  
  for(i=0;i<n;i++)
    FREE_REF_STRING(BUT(button).action[i]);

  FREE(BUT(button).action);
  KILL_IM(im_data,button,im_orig);
  KILL_IM(im_data,button,im_normal);
  KILL_IM(im_data,button,im_snap);
  FREE_PIX(im_data,button,icon_mask);
  FREE_PIX(im_data,button,bg_mask);

  if (button->win) {
    XDeleteContext(display,button->win,xid_context);
    XDestroyWindow(display,button->win);
  }

  if (IS_PIXMAP(button->tooltip_bg))
    FREE_REF_STRING(GET_PIXMAP(button->tooltip_bg));
}


#define PER_KILL_IM(NAME)\
 do { if (PERIODIC(gad).NAME) {\
  trace(Imlib_kill_image(im_data,PERIODIC(gad).NAME));\
  PERIODIC(gad).NAME=NULL; } } while(0)

/* free all the buttons and the pages + icon paths + key bindings
 */
void free_gadgets() {
  int i,j;
  debug( int nbuttons=0 );

  assert((npages==0)==(pages==NULL),"Inconsistent data");
  if (pages) {
    /* the first one always point to the array
     * of all the buttons ; IF any button is defined
     */

    for(i=0;i<npages;i++) {
      debug( nbuttons+=NBUTTONS(i) );
      assert((nbuttons==0) == (pages[i].buttons==NULL),
	     "Inconsistent data");
      for(j=0;j<NBUTTONS(i);j++)
	freeButtonContent(&pages[i].buttons[j]);
    }

    FREE(pages[0].buttons);
    debugPrint(("freed %d buttons\n",nbuttons));
    disalloc(pages);
    npages=0;
    pages=NULL;
  }

  assert(npaths>=0,"Path number is negative!");
  assert((icon_paths==NULL)==(npaths==0),"Inconsistent data");
  if (icon_paths>0) {
    for(i=0;i<npaths;i++) {
      assert(icon_paths[i]!=NULL,"A path is null");
      disalloc(icon_paths[i]);
    }
    disalloc(icon_paths);
    npaths=0;
    longest_path_size=0;
    icon_paths=NULL;
  }

  assert(nkeybindings>=0,"Key binding number is negative!");
  assert((keybindings==NULL)==(nkeybindings==0),
	 "Inconsistent data");
  if (keybindings>0) {
    for(i=0;i<nkeybindings;i++) {
      assert(keybindings[i]!=NULL,"A key binding is null");
      disalloc(keybindings[i]);
    }
    disalloc(keybindings);
    nkeybindings=0;
    keybindings=NULL;
  }

  assert(nb_init>=0,"Init number is negative");
  assert((nb_init==0)==(init_commands==NULL),
	 "Inconsistent data");
  if (nb_init>0) {
    for(i=0;i<nb_init;i++) {
      assert(init_commands[i].init_command!=NULL,
	     "A command is null");
      disalloc(init_commands[i].init_command);
    }
    disalloc(init_commands);
    init_commands=NULL;
    nb_init=0;
  }

  FREE(jobs);
  jobs=NULL;
  n_jobs=0;

  free_fonts();
  free_tooltip();

  assert((extGadgets==NULL)==(nb_extGadgets==0),"Inconsistent data");
  for(i=0;i<nb_extGadgets;i++)
    if (IS_PERIODIC(extGadgets+i)) {
      Gadget *gad=extGadgets+i;
      if (IS_PIXMAP(gad->bg))
	FREE_REF_STRING(GET_PIXMAP(gad->bg));
      if (IS_PIXMAP(gad->tooltip_bg))
	FREE_REF_STRING(GET_PIXMAP(gad->tooltip_bg));
      if (gad->win) {
	XDeleteContext(display,gad->win,xid_context);
	XDestroyWindow(display,gad->win);
      }
      PER_KILL_IM(im_orig);
      PER_KILL_IM(im_normal);
      PERIODIC(gad).freeGadget(extGadgets+i);
    }
  FREE(extGadgets);
  extGadgets=NULL;
  nb_extGadgets=0;

  checkFreed();
}


void do_nothing(Gadget *gad) { }
int return_false(Gadget *gad,fd_set* fdset) { return 0; }
void do_nothing2(Gadget *gad,int *nfd,fd_set* fdset) { }
