/* gadget.h
 *
 * Part of FancyLauncher
 * the gadget type: button
 *
 * author: Alexandre David
 * e-mail: adavid@docs.uu.se
 */

#ifndef GADGET_H
#define GADGET_H

#include <Imlib.h>
#include "font.h"
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/* ------------------------------------------------------------- *
 * The background
 * an address is always even, so we
 * use the 1st bit as a flag:
 * 0: address -> pixmap
 * 1: color
 * ------------------------------------------------------------- */

typedef union {
  RefString pixmapRefString;
  unsigned int colorCode;
} Background;


#define SET_PIXMAP(bg,pixmap) \
 do {\
   SET_REF_STRING((bg).pixmapRefString,pixmap);\
   assert(((bg).colorCode&1)==0,"Even address assertion");\
 } while(0)

#define SET_COLOR(bg,color) (bg).colorCode=((color)<<1)|1

#define IS_PIXMAP(bg) (((bg).colorCode&1)^1 && (bg).pixmapRefString)
#define IS_COLOR(bg)   ((bg).colorCode&1)

#define GET_COLOR(bg) ((bg).colorCode>>1)
#define GET_PIXMAP(bg) (bg).pixmapRefString

/* special case */
#define IS_TRANSPARENT(bg) ((bg).pixmapRefString==NULL)
#define SET_TRANSPARENT(bg) ((bg).pixmapRefString=NULL)



/* --------------------------------------------------------- *
 * Options: sticky, below, snap are common to all types ;
 * mapped is common, it is used to know if the window is mapped
 * moving is common, it says if the window is moving
 * it is better to have a bit per option, but
 * this is worse for implementation and readability
 * and we do not loose so much memory
 * --------------------------------------------------------- */

typedef enum {
  I_STICKY=0,   I_BELOW,    I_SNAP,        I_MAPPED,  I_MOVING,

  I_TOOLTIPS,

  I_RAISED,     I_ETCHED,   I_TRANSIENT,   I_RESTORE,
  I_BLEND_ICON, I_BLEND_BG, I_BLEND_FOCUS, I_PAGE,    I_NACTIONS
} ButtonOpt;

#define NB_OPT 15 /* 15 => +1 (type) = 4*sizeof(int) */
#define COMMON_NB_OPT (I_TOOLTIPS+1)

/* generic type that have common options */

typedef struct {
  unsigned char type;
  unsigned char opt[NB_OPT];
} AnyGadget;


/* in order to check common options */
typedef struct {
  unsigned int opt[4];
} RawInt;


/* option resource type */
typedef struct {
  unsigned int parseCode;
  char *name;
  char *comment;
} OptionRC;

/* option resources for buttons */
extern const OptionRC buttonRC[NB_OPT];


/* ------------------------------------------------------------------- *
 * Button type: compatible with AnyType
 * button appearences:
 * - normal : icon+bg blent with root
 * - focused: not blent lighter with root
 * - pressed: not blent darker with root
 * when the icon moves, we need icon+bg not blent
 * to blend the image with root.
 * when focused/pressed we take transient snapshot
 * of the window on the screen.
 * we want to avoid flicker when switching normal/focused/pressed
 * => im_orig = icon+bg blent together
 * => im_normal = im_orig blent with root
 * => im_snap = transient snap (avoid flicker lighter/darker)
 * masks needed: icon+bg
 *-------------------------------------------------------------------- */

typedef struct {
  unsigned char type;
  unsigned char opt[NB_OPT];

  RefString icon;
  RefString *action; /* even: action ; odd: tooltip */

  ImlibImage *im_orig,*im_normal,*im_snap;
  Pixmap icon_mask,bg_mask;

} ButtonGadget;


#define KILL_IM(ID,BUTTON,IM) \
 do { if (BUT(BUTTON).IM) trace( Imlib_kill_image(ID,BUT(BUTTON).IM) ); } while(0)

#define FREE_PIX(ID,BUTTON,PIX) \
 do { if (BUT(BUTTON).PIX) trace( Imlib_free_pixmap(ID,BUT(BUTTON).PIX) ); } while(0)

/* ------------------------------------------------- *
 * a periodic gadget is active and runs periodically *
 * ------------------------------------------------- */

struct t_Gadget; /* forward declaration */

typedef struct {
  unsigned char type;
  unsigned char opt[NB_OPT];

  int win_width,win_height;
  ImlibImage *im_orig,*im_normal;
  void *data;
  double update_period;

  void (*saveGadget)(FILE *,struct t_Gadget *);
  double (*updateGadget)(struct t_Gadget *,int *nfd,fd_set *,int); /* returns period to wait */
  void (*pressedGadget)(struct t_Gadget *); /* action */
  void (*releasedGadget)(struct t_Gadget *);
  void (*freeGadget)(struct t_Gadget *);    /* cleanup */
  char* (*tooltipGadget)(struct t_Gadget *);/* returns static tooltip string */
  void (*resizedGadget)(struct t_Gadget *);
  void (*movedGadget)(struct t_Gadget *);
  void (*enteredGadget)(struct t_Gadget *);
  void (*leftGadget)(struct t_Gadget *);
  int (*calledGadget)(struct t_Gadget *,fd_set *);
  void (*addFD)(struct t_Gadget *,int *nfd,fd_set *);

} PeriodicGadget;

const char *labelnames[8]; /* used for font pos/color in periodic gadgets */

typedef struct {
  double time;
  struct t_Gadget *gad;
} TimedPeriodicAction;


extern TimedPeriodicAction *jobs;
extern int n_jobs;


/* -------------------- *
 * Gadgets
 * -------------------- */

typedef union {
  AnyGadget any;
  ButtonGadget button;
  PeriodicGadget periodic;

  RawInt raw; /* special: to check common options */

  /* open for extensions here */
} GadgetUnion;

typedef enum { BUTTON_TYPE=0,PERIODIC_TYPE=1,NONE=2 } GadgetType;

typedef struct t_Gadget {
  Window win;
  int x,y,width,height;

  Background bg;
  GadgetUnion gadget;

  FontInfo *tooltip_font;
  Background tooltip_bg;
  int tooltip_bw,tooltip_bh;
  int gradColor[6];
} Gadget;


#define ANY(gad)                      (gad)->gadget.any
#define BUT(gad)                      (gad)->gadget.button
#define PERIODIC(gad)                 (gad)->gadget.periodic
#define IS_BUTTON(gad)                (ANY(gad).type==BUTTON_TYPE)
#define IS_PERIODIC(gad)              (ANY(gad).type==PERIODIC_TYPE)

#define HAS_PIXMAP_BG(gad)            IS_PIXMAP((gad)->bg)
#define HAS_TRANSPARENT_BG(gad)       IS_TRANSPARENT((gad)->bg)
#define HAS_SOLID_COLOR_BG(gad)       IS_COLOR((gad)->bg)

#define IS_STICKY(gad)                ANY(gad).opt[I_STICKY]
#define IS_BELOW(gad)                 ANY(gad).opt[I_BELOW]
#define IS_SNAPPED(gad)               ANY(gad).opt[I_SNAP]
#define IS_MAPPED(gad)                ANY(gad).opt[I_MAPPED]
#define IS_MOVING(gad)                ANY(gad).opt[I_MOVING]
#define SHOWS_TOOLTIP(gad)            ANY(gad).opt[I_TOOLTIPS]

#define IS_BORDER_RAISED(but)         BUT(but).opt[I_RAISED]
#define IS_BORDER_ETCHED(but)         BUT(but).opt[I_ETCHED]
#define IS_BORDER_TRANSIENT(but)      BUT(but).opt[I_TRANSIENT]
#define FOCUS_RESTORES_ICON(but)      BUT(but).opt[I_RESTORE]
#define BLEND_ICON_BG(but)            BUT(but).opt[I_BLEND_ICON]
#define BLEND_BG_ROOT(but)            BUT(but).opt[I_BLEND_BG]
#define BLEND_WHEN_FOCUSED(but)       BUT(but).opt[I_BLEND_FOCUS]
#define PAGEOF(but)                   BUT(but).opt[I_PAGE]
#define NACTIONS(but)                 BUT(but).opt[I_NACTIONS]

#define HAS_BORDER(but) \
 (IS_BORDER_RAISED(but) | IS_BORDER_ETCHED(but))


typedef struct {
  int nbuttons;
  Gadget *buttons;
} ButtonPage;

#define NBUTTONS(pnr)   pages[pnr].nbuttons

/* similar coding as RefString, the int is the keycode this time */
typedef char* KeyBinding;

#define KEYCODE(kb)    (*((unsigned int*)(kb)))
#define KEYCOMMAND(kb) CHARS(kb)


/* ---------------------- *
 * type for init commands *
 * ---------------------- */

typedef struct {
  char *init_command;
  int button_id;
} InitCommand;


/* declaration of core variables */

extern int npages,current_page;
extern ButtonPage *pages;

extern int npaths,longest_path_size;
extern char **icon_paths;

extern int darker,lighter;

extern int nkeybindings;
extern KeyBinding *keybindings;

extern InitCommand *init_commands;
extern int nb_init;

extern ImlibImage **tooltip_icons;
extern int nb_tooltip_icons;

extern Gadget *extGadgets;
extern int nb_extGadgets;


/* returns the buttons in page p ; n will contain the
 * number of buttons in this page
 */
Gadget* getButtons(const int p,int *n);


/* returns the action string number i of a button ; or NULL
 */
char* getAction(const Gadget* button,const int i);


/* returns the action string bound to a key
 */
char* getKeyAction(unsigned int code);


/* returns the tooltip string number i of a button ; or NULL
 */
char* getTooltip(const Gadget* button,const int i);


typedef enum { NORMAL,FOCUSED,PRESSED } ButtonState;

/* paint the button ; force reinitialization of the pixmaps
 */
void paintReInitButton(ImlibData *id,Gadget *button);


/* paint the button ; initialize the pixmaps
 */
void paintInitButton(ImlibData *id,Gadget *button);


/* paint a button in a given state
 */
void paintButton(ImlibData *id,Gadget *button,ButtonState state);


/* free the content of a button, not the button itself
 */
void freeButtonContent(Gadget *button);


/* free internal structures
 */
void free_gadgets();

ImlibImage* getPixmapBg(ImlibData *id,const Gadget* button);


void do_nothing(Gadget *gad);
int return_false(Gadget *gad,fd_set *fdset);
void do_nothing2(Gadget *gad,int *nfd,fd_set *fdset);

#endif


