/* load_save
 * load and save configuration of the buttons
 * part of FancyLauncher
 *
 * author: Alexandre David
 * e-mail: adavid@docs.uu.se
 * http://www.docs.uu.se/~adavid
 * license: GPL
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/resource.h>
#include <unistd.h>
#include <sys/types.h>

#include "debug.h"
#include "utils.h"
#include "gadget.h"
#include "load_save.h"
#include "pixmaps.h"
#include "codes.h"
#include "Xcore.h"
#include "font.h"
#include "tooltips.h"
#include "clock.h"
#include "mailcheck.h"

char *crypt(const char *key, const char *salt);


static char *configFile;

static const char *color_names[]={
  "tooltip_red_start","tooltip_red_end",
  "tooltip_green_start","tooltip_green_end",
  "tooltip_blue_start","tooltip_blue_end"};

/* states for parsing
 */
enum { NOWHERE,
       COMMON_FOUND, IN_COMMON,
       BUTTON_FOUND, IN_BUTTON,
       DCLOCK_FOUND, IN_DCLOCK,
       DCLOCK_LABEL_FOUND, IN_DCLOCK_LABEL,
       MAILCHECK_FOUND, IN_MAILCHECK };


/* try to find a shared string
 */
static RefString findRefString(const Gadget *button,const int n,const RefString s) {
  int i;

  pushArgs(64,(BUF,"%s,%d,%s",PTR2STR(button),n,PTR2STR(s)));
  assert(n==0 || button!=NULL,"Inconsistent button data");
  assert(s!=NULL,"Lookup string null");
  assert(n>=0,"Bad argument value");
  for(i=0;i<n;i++) {
    int na,j;
    assert(IS_BUTTON(button+i),"Button type expected");

    /* try the background pixmap */
    if (HAS_PIXMAP_BG(button+i) &&
	!strcmp(CHARS(GET_PIXMAP(button[i].bg)),CHARS(s))) {

      debugPrint((GREEN"****"BLACK" findRefString>"GREEN"%s"BLACK"\n",CHARS(s)));
      return GET_PIXMAP(button[i].bg);
    }

    /* try the tooltip background pixmap */
    if (IS_PIXMAP(button[i].tooltip_bg) &&
	!strcmp(CHARS(GET_PIXMAP(button[i].tooltip_bg)),CHARS(s))) {

      debugPrint((GREEN"****"BLACK" findRefString>"GREEN"%s"BLACK"\n",CHARS(s)));
      return GET_PIXMAP(button[i].tooltip_bg);
    }

    /* try the icon pixmap */
    if (BUT(button+i).icon &&
	!strcmp(CHARS(BUT(button+i).icon),CHARS(s))) {

      debugPrint((GREEN"****"BLACK" findRefString>"GREEN"%s"BLACK"\n",CHARS(s)));
      return BUT(button+i).icon;
    }

    na=NACTIONS(button+i);

    /* na == 0 <=> action == 0 */
    assert((na==0) == (BUT(button+i).action==NULL),
	   "Inconsistent data");

    /* try actions/tooltips */
    if (BUT(button+i).action) {
      debug( int countthem=0 );
      for(j=0;j<2*na;j++) {
	debug(if (BUT(button+i).action[j]) countthem++);	  
	if (BUT(button+i).action[j] &&
	    !strcmp(CHARS(BUT(button+i).action[j]),CHARS(s)))

	  return BUT(button+i).action[j];
      }
      assert(countthem>0,"No actions and tooltips though the pointer is not null");
      /* actions/tooltips => one is not null */
    }

  }

  /* not found */
  return NULL;
}


static ClockData* get_clock_data(Gadget *gad) {
  ClockData *ckdata;
  ckdata= gad ? (ClockData*)PERIODIC(gad).data : NULL ;
  if (ckdata==NULL) {
    trace(ckdata=(ClockData*)safeAllocPtr(alloc(sizeof(ClockData))));
    if (gad)
      PERIODIC(gad).data=ckdata;
    ckdata->n=0;
    ckdata->label=NULL;
  }
  return ckdata;
}


static PopData* getPopData() {
  PopData* dat;
  trace(dat=(PopData*)safeAllocPtr(alloc(sizeof(PopData))));
  memset(dat,0,sizeof(PopData));
  return dat;
}


/* copy general information between gadgets
 */
static void copyGadget(const Gadget *src,Gadget *dst) {
  pushArgs(32,(BUF,"%s,%s",PTR2STR(src),PTR2STR(dst)));
  assert(src!=NULL,"Source button is null");
  assert(dst!=NULL,"Destination button is null");

  memset(dst,0,sizeof(Gadget));
  dst->x=src->x;
  dst->y=src->y;
  dst->width=src->width;
  dst->height=src->height;
  if (!IS_COLOR(src->bg))
    SET_REF_STRING(GET_PIXMAP(dst->bg),GET_PIXMAP(src->bg));
  else
    dst->bg.colorCode=src->bg.colorCode;

  memcpy(dst->gadget.any.opt,src->gadget.any.opt,NB_OPT);

  dst->tooltip_font=src->tooltip_font;
  if (!IS_COLOR(src->tooltip_bg))
    SET_REF_STRING(GET_PIXMAP(dst->tooltip_bg),GET_PIXMAP(src->tooltip_bg));
  else
    dst->tooltip_bg.colorCode=src->tooltip_bg.colorCode;

  dst->tooltip_bw=src->tooltip_bw;
  dst->tooltip_bh=src->tooltip_bh;
  memcpy(dst->gradColor,src->gradColor,6*sizeof(int));
}


/* copy a button ; treat the RefString as they should
 */
static void copyButton(const Gadget *src,Gadget *dst) {
  int i;

  pushArgs(32,(BUF,"%s,%s",PTR2STR(src),PTR2STR(dst)));
  assert(src!=NULL,"Source button is null");
  assert(dst!=NULL,"Destination button is null");
  assert(src==NULL || IS_BUTTON(src),"Expected type : button");

  /* naction == 0 <=> action == NULL */
  assert((NACTIONS(src)==0) == (BUT(src).action==NULL),
	 "Inconsistent data");

  *dst=*src;

  if (!HAS_SOLID_COLOR_BG(src))
    SET_REF_STRING(GET_PIXMAP(dst->bg),GET_PIXMAP(src->bg));

  SET_REF_STRING(BUT(dst).icon,BUT(src).icon);

  if (BUT(src).action)
    trace( BUT(dst).action=(RefString*)safeAllocPtr(alloc(NACTIONS(src)*2*sizeof(RefString))) );

  for(i=0;i<2*NACTIONS(src);i++)
    SET_REF_STRING(BUT(dst).action[i],BUT(src).action[i]);

  if (!IS_COLOR(src->tooltip_bg))
    SET_REF_STRING(GET_PIXMAP(dst->tooltip_bg),GET_PIXMAP(src->tooltip_bg));

  dst->win=0;
  BUT(dst).im_orig=0;
  BUT(dst).im_normal=0;
  BUT(dst).im_snap=0;  
}


/* parse the configuration file
 * and build internal structures
 */
void configParse() {
  FILE *f;
  char buffer[1024];
  RefString arg=NULL;
  int state=0,line=0,col,i,n_all=0,caseX,*intptr,token;
  Gadget default_button;
  Gadget *thebutton=&default_button,*all_buttons=NULL;
  Background *bg_ptr;
  ClockData *clockData=NULL;

  pushMe();
  memset(&default_button,0,sizeof(Gadget));

  default_button.width=32;
  default_button.height=32;
  default_button.tooltip_bw=16;
  default_button.tooltip_bh=16;
  SET_COLOR(default_button.tooltip_bg,0x707070);

  /* everything explicit, the compiler will optimize this */
  default_button.gadget.button.type=BUTTON_TYPE;
  default_button.gadget.button.opt[I_STICKY]=1;
  default_button.gadget.button.opt[I_BELOW]=1;
  default_button.gadget.button.opt[I_SNAP]=1;
  default_button.gadget.button.opt[I_MAPPED]=0;
  default_button.gadget.button.opt[I_MOVING]=0;
  default_button.gadget.button.opt[I_RAISED]=0;
  default_button.gadget.button.opt[I_ETCHED]=1;
  default_button.gadget.button.opt[I_TRANSIENT]=1;
  default_button.gadget.button.opt[I_TOOLTIPS]=1;
  default_button.gadget.button.opt[I_RESTORE]=1;
  default_button.gadget.button.opt[I_BLEND_ICON]=128;
  default_button.gadget.button.opt[I_BLEND_BG]=128;
  default_button.gadget.button.opt[I_BLEND_FOCUS]=255;
  default_button.gadget.button.opt[I_PAGE]=0;
  default_button.gadget.button.opt[I_NACTIONS]=0;

  trace( free_gadgets() );

  f=fopen(configFile,"r");
  if (!f) {
    fprintf(stderr,
	    "Couldn't open configuration file %s, taking default /usr/local/FancyLauncher/config\n",
	    configFile);
    f=fopen("/usr/local/FancyLauncher/data/config","r");
  }
  if (!f) {
    fprintf(stderr,
	    "Could not open default configuration file, please check your installation\n");
    exit(1);
  }

  while(fgets(buffer,sizeof(buffer),f)) {
    line++;
    col=0;
    while(jumpBlank(buffer,&col) && buffer[col]!='#') {

      pushMe();
      updateMem();
      switch((token=readToken(buffer,&col))) {
	
      case COMMON:
	checkMem();
	checkState(NOWHERE,"}");
	state=COMMON_FOUND;
	thebutton=&default_button;
	break;

      case BUTTON:
	checkMem();
	checkState(NOWHERE,"}");
	state=BUTTON_FOUND;
	n_all++;
	trace( all_buttons=(Gadget*)safeAllocPtr(ralloc(all_buttons,n_all*sizeof(Gadget))) );
	thebutton=all_buttons+n_all-1;
	trace( copyButton(&default_button,thebutton) );
	break;
	


      case DIGITAL_CLOCK:
	checkMem();
	checkState(NOWHERE,"}");
	state=DCLOCK_FOUND;
	nb_extGadgets++;
	trace( extGadgets=(Gadget*)safeAllocPtr(ralloc(extGadgets,nb_extGadgets*sizeof(Gadget))) );
	thebutton=extGadgets+nb_extGadgets-1;
	trace( copyGadget(&default_button,thebutton) );
	PERIODIC(thebutton).win_width=thebutton->width;
	PERIODIC(thebutton).win_height=thebutton->height;
	ANY(thebutton).type=PERIODIC_TYPE;
	PERIODIC(thebutton).update_period=1.0;
	PERIODIC(thebutton).saveGadget=saveClockGadget;
	PERIODIC(thebutton).updateGadget=updateClockGadget;
	PERIODIC(thebutton).pressedGadget=do_nothing;
	PERIODIC(thebutton).releasedGadget=do_nothing;
	PERIODIC(thebutton).freeGadget=freeClockGadget;
	PERIODIC(thebutton).tooltipGadget=tooltipClockGadget;
	PERIODIC(thebutton).resizedGadget=do_nothing;
	PERIODIC(thebutton).movedGadget=do_nothing;
	PERIODIC(thebutton).enteredGadget=enteredClockGadget;
	PERIODIC(thebutton).leftGadget=leftClockGadget;
	PERIODIC(thebutton).calledGadget=return_false;
	PERIODIC(thebutton).data=clockData=get_clock_data(NULL);
	PERIODIC(thebutton).addFD=do_nothing2;
	break;

      case MAIL_CHECK:
	checkMem();
	checkState(NOWHERE,"}");
	state=MAILCHECK_FOUND;
	nb_extGadgets++;
	trace( extGadgets=(Gadget*)safeAllocPtr(ralloc(extGadgets,nb_extGadgets*sizeof(Gadget))) );
	thebutton=extGadgets+nb_extGadgets-1;
	trace( copyGadget(&default_button,thebutton) );
	PERIODIC(thebutton).win_width=thebutton->width;
	PERIODIC(thebutton).win_height=thebutton->height;
	ANY(thebutton).type=PERIODIC_TYPE;
	PERIODIC(thebutton).update_period=1.0;
	PERIODIC(thebutton).saveGadget=saveMailGadget;
	PERIODIC(thebutton).updateGadget=updateMailGadget;
	PERIODIC(thebutton).pressedGadget=do_nothing;
	PERIODIC(thebutton).releasedGadget=do_nothing;
	PERIODIC(thebutton).freeGadget=freeMailGadget;
	PERIODIC(thebutton).tooltipGadget=tooltipMailGadget;
	PERIODIC(thebutton).resizedGadget=do_nothing;
	PERIODIC(thebutton).movedGadget=do_nothing;
	PERIODIC(thebutton).enteredGadget=enteredMailGadget;
	PERIODIC(thebutton).leftGadget=leftMailGadget;
	PERIODIC(thebutton).calledGadget=calledMailGadget;
	PERIODIC(thebutton).data=getPopData();
	PERIODIC(thebutton).addFD=addMailFD;
	break;

      case MAIL:
	checkMem();
	if CHECK_COND(state==IN_MAILCHECK,"mail") {
	  if (IS_PERIODIC(thebutton) && PERIODIC(thebutton).data!=NULL) {
	    if (checkArg(buffer,&col,line,"Command")) {
	      PopData *pop=(PopData*)PERIODIC(thebutton).data;
	      pop->command=readChars(buffer,&col,0);
	    }
	  }
	}
	break;

      case CODED_PASSWD:
	caseX=0;
	arg="coded_passwd";
	goto pop_passwd_read;

      case PASSWD:
	caseX=1; /* need to code */
	arg="passwd";
      pop_passwd_read:
	checkMem();
	if CHECK_COND(state==IN_MAILCHECK,"passwd") {
	  if (IS_PERIODIC(thebutton) && PERIODIC(thebutton).data!=NULL) {
	    if (checkArg(buffer,&col,line,"Crypt password")) {
	      char *crypt_pass;
	      checkMem();
	      crypt_pass=buffer+col;
	      trace(col+=jumpWordNull(crypt_pass)-crypt_pass);
	      if (checkArg(buffer,&col,line,"User name")) {
		char *name=buffer+col;
		checkMem();
		trace(col+=jumpWordNull(name)-name);
		if (checkArg(buffer,&col,line,"Pop server")) {
		  char *server=buffer+col;
		  checkMem();
		  trace(col+=jumpWordNull(server)-server);
		  if (checkArg(buffer,&col,line,"Pop password")) {
		    char *pop_pass=readChars(buffer,&col,0);
		    PopData *pop=(PopData*)PERIODIC(thebutton).data;

		    if (caseX)
		      trace(code_passwd(crypt_pass,pop_pass));
		    i=pop->nb_pop;
		    debugPrint(("Nb pop: %d\n",pop->nb_pop));
		    pop->nb_pop++;
		    trace(pop->pop_info=(PopInfo*)safeAllocPtr(ralloc(pop->pop_info,pop->nb_pop*sizeof(PopInfo))));
		    memset(pop->pop_info+i,0,sizeof(PopInfo));
		    trace(pop->pop_info[i].crypt_pass=(char*)safeAllocPtr(strdup(crypt_pass)));
		    addStringBlock(pop->pop_info[i].crypt_pass);
		    trace(pop->pop_info[i].username=(char*)safeAllocPtr(strdup(name)));
		    addStringBlock(pop->pop_info[i].username);
		    trace(pop->pop_info[i].server=(char*)safeAllocPtr(strdup(server)));
		    addStringBlock(pop->pop_info[i].server);
		    trace(pop->pop_info[i].passwd=(char*)pop_pass);
		    pop->pop_info[i].fd=-1;
		  }
		}
	      }
	    }
	  } else {
	    fprintf(stderr,debug(YELLOW)
		    "Parse error at line %d, column %d!!"
		    debug(BLACK)"\n",line,col);
	    state=NOWHERE;
	  }
	}
	break;

      case LABEL:
	checkMem();
	if CHECK_COND(state==IN_DCLOCK,"label") {
	  if (IS_PERIODIC(thebutton)) {
	    ClockData *ckdata=clockData=get_clock_data(thebutton);
	    ckdata->n++;
	    trace(ckdata->label=(ClockLabel*)safeAllocPtr(ralloc(ckdata->label,ckdata->n*sizeof(ClockLabel))));
	    memset(ckdata->label+ckdata->n-1,0,sizeof(ClockLabel));
	    state=DCLOCK_LABEL_FOUND;
	  } else {
	    fprintf(stderr,debug(YELLOW)
		    "Parse error at line %d, column %d!"
		    debug(BLACK)"\n",line,col);
	    state=NOWHERE;
	  }
	}
	break;


      case L_PAR:
	checkMem();
	if CHECK_COND(state==COMMON_FOUND || state==BUTTON_FOUND ||
		      state==DCLOCK_FOUND || state==DCLOCK_LABEL_FOUND ||
		      state==MAILCHECK_FOUND,"{")
	  state++;
	break;
	
      case R_PAR:
	checkMem();
	switch(state) {
	case IN_DCLOCK_LABEL: state=IN_DCLOCK; break;
	case IN_BUTTON:
	case IN_COMMON:
	case IN_MAILCHECK:
	case IN_DCLOCK: state=NOWHERE; clockData=NULL; break;
	default:
	  fprintf(stderr,debug(YELLOW)
		  "Unexpected \"}\" at line %d, column %d\n"
		  debug(BLACK),line,col+1);
	  state=NOWHERE;
	}
	break;

      case DARKER:
	arg="darker";
	caseX=0;
	goto darker_or_lighter;

      case LIGHTER:
	arg="lighter";
	caseX=1;
      darker_or_lighter:
	checkMem();
	if CHECK_COND(state==IN_COMMON,arg) {
	  assert(thebutton!=NULL,"Current button is null!");
	  trace( readIntArg(buffer,&col,line,&i) );
	  trace( i=checkPercent(i,line,col) );
	  if (caseX)
	    lighter=(i*255)/100+255;
	  else
	    darker=((100-i)*255)/100;
	}
	break;

      case TOOLTIP_WAIT:
	checkMem();
	if CHECK_COND(state==IN_COMMON,"tooltip_wait") {
	  trace( readDoubleArg(buffer,&col,line,&tooltip_period) );
	  if (tooltip_period<0.0) {
	    fprintf(stderr,debug(YELLOW)
		      "Tooltip value %g rejected, negative, 0.0 taken instead"
		      debug(BLACK)"\n",tooltip_period);
	    tooltip_period=0.0;
	  }
	}
	break;

      case UPDATE:
	checkMem();
	if CHECK_COND(state==IN_DCLOCK||state==IN_MAILCHECK,"update") {
	  if (IS_PERIODIC(thebutton)) {
	    trace( readDoubleArg(buffer,&col,line,&PERIODIC(thebutton).update_period) );
	    if (PERIODIC(thebutton).update_period<0.1) {
	      fprintf(stderr,debug(YELLOW)
		      "Update value %g rejected, too small, 1.0 taken instead"
		      debug(BLACK)"\n",PERIODIC(thebutton).update_period);
	      PERIODIC(thebutton).update_period=1.0;
	    }
	  } else {
	    fprintf(stderr,debug(YELLOW)
		    "Parse error at line %d, column %d!"debug(BLACK)"\n",
		    line,col);
	    state=NOWHERE;
	  }
	}
	break;

      case RESIZED_WIDTH:
	checkMem();
	if CHECK_COND(state==IN_DCLOCK,"resized_width") {
	  if (IS_PERIODIC(thebutton)) {
	    trace( readIntArg(buffer,&col,line,&PERIODIC(thebutton).win_width) );
	  } else {
	    fprintf(stderr,debug(YELLOW)
		    "Parse error at line %d, column %d!"debug(BLACK)"\n",
		    line,col);
	    state=NOWHERE;
	  }
	}
	break;

      case RESIZED_HEIGHT:
	checkMem();
	if CHECK_COND(state==IN_DCLOCK,"resized_heigh") {
	  if (IS_PERIODIC(thebutton)) {
	    trace( readIntArg(buffer,&col,line,&PERIODIC(thebutton).win_height) );
	  } else {
	    fprintf(stderr,debug(YELLOW)
		    "Parse error at line %d, column %d!"debug(BLACK)"\n",
		    line,col);
	    state=NOWHERE;
	  }
	}
	break;

      case COLOR_RED_START:
	arg="color_red_start";
	i=0;
	goto read_label_int;

      case COLOR_RED_END:
	arg="color_red_end";
	i=1;
	goto read_label_int;

      case COLOR_GREEN_START:
	arg="color_green_start";
	i=2;
	goto read_label_int;

      case COLOR_GREEN_END:
	arg="color_green_end";
	i=3;
	goto read_label_int;

      case COLOR_BLUE_START:
	arg="color_blue_start";
	i=4;
	goto read_label_int;

      case COLOR_BLUE_END:
	arg="color_blue_end";
	i=5;
	goto read_label_int;

      case POSX:
	arg="posx";
	i=6;
	goto read_label_int;

      case POSY:
	arg="posy";
	i=7;
      read_label_int:
	checkMem();
	assert(thebutton!=NULL,"Null current button");
	if (state==IN_DCLOCK_LABEL) {
	  if (IS_PERIODIC(thebutton) && clockData && clockData->label) {
	    assert(clockData->n>0,"In a label and no data!??");
	    assert(i>=0 && i<8,"Index overflow, kaboum");
	    trace( readIntArg(buffer,&col,line,&clockData->label[clockData->n-1].color[i]) );
	  } else {
	    fprintf(stderr,debug(YELLOW)
		    "Parse error at line %d, column %d: %s %s!"debug(BLACK)"\n",
		    line,col,
		    clockData?"ok":"null",
		    clockData && clockData->label?"ok":"null");
	    state=NOWHERE;
	  }
	} else if (state==IN_MAILCHECK) {
	  if (IS_PERIODIC(thebutton)) {
	    PopData *pop=(PopData*)PERIODIC(thebutton).data;
	    if (pop) {
	      assert(i>=0 && i<8,"Index overflow, kaboum");
	      trace( readIntArg(buffer,&col,line,&pop->color[i]) );
	    }
	  }
	} else
	  fprintf(stderr,debug(YELLOW)"Unexpected \"%s\" at line %d, column %d\n"
		  debug(BLACK),arg,line,col+1);
	break;

      case TOOLTIP_RED_START:
	arg="tooltip_color_start";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->gradColor[0];
	goto read_an_int;

      case TOOLTIP_RED_END:
	arg="tooltip_color_end";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->gradColor[1];
	goto read_an_int;

       case TOOLTIP_GREEN_START:
	arg="tooltip_color_start";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->gradColor[2];
	goto read_an_int;

      case TOOLTIP_GREEN_END:
	arg="tooltip_color_end";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->gradColor[3];
	goto read_an_int;      

      case TOOLTIP_BLUE_START:
	arg="tooltip_color_start";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->gradColor[4];
	goto read_an_int;

      case TOOLTIP_BLUE_END:
	arg="tooltip_color_end";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->gradColor[5];
	goto read_an_int;

      case TOOLTIP_BORDER_WIDTH:
	arg="tooltip_border_width";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->tooltip_bw;
	goto read_an_int;

      case TOOLTIP_BORDER_HEIGHT:
	arg="tooltip_border_height";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->tooltip_bh;
	goto read_an_int;

      case WIDTH:
	arg="width";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->width;
	goto read_an_int;

      case HEIGHT:
	arg="height";
	assert(thebutton!=NULL,"Null current button");
	intptr=&thebutton->height;
      read_an_int:
	checkMem();
	if CHECK_COND(state==IN_BUTTON || state==IN_COMMON ||
		      state==IN_DCLOCK || state==IN_MAILCHECK,
		      arg) {
	  assert(thebutton!=NULL,"Current button is null!");
	  trace( readIntArg(buffer,&col,line,intptr) );
	}
	break;
	
      case NEW_MAIL_BG:
	if CHECK_COND(state==IN_MAILCHECK,"new_mail_bg") {
	  PopData *pop=(PopData*)PERIODIC(thebutton).data;
	  if (pop) {
	    arg="new_mail_bg";
	    bg_ptr=&pop->new_bg;
	    goto read_background;
	  }
	}
	break;

      case ERROR_MAIL_BG:
	if CHECK_COND(state==IN_MAILCHECK,"error_mail_bg") {
	  PopData *pop=(PopData*)PERIODIC(thebutton).data;
	  if (pop) {
	    arg="new_mail_bg";
	    bg_ptr=&pop->error_bg;
	    goto read_background;
	  }
	}
	break;

      case TOOLTIP_BACKGROUND:
	arg="tooltip_background";
	assert(thebutton!=NULL,"Null current button");
	bg_ptr=&thebutton->tooltip_bg;
	goto read_background;

      case BACKGROUND:
	arg="background";
	assert(thebutton!=NULL,"Null current button");
	bg_ptr=&thebutton->bg;
      read_background:
	checkMem();
	if CHECK_COND(state==IN_BUTTON || state==IN_COMMON ||
		      state==IN_DCLOCK || state==IN_MAILCHECK,arg) {
	  if ( checkArg(buffer,&col,line,
			"Color 0xBBGGRR or \"transparent\" of \"none\" or pixmap")) {

	    checkMem();
	    trace( arg=readString(buffer,&col) );
	    assert(arg!=NULL,"Null argument read");

	    if (!strcmp(CHARS(arg),"transparent") ||
		!strcmp(CHARS(arg),"none")) {
	      
	      SET_TRANSPARENT(*bg_ptr);
	      
	    } else if (!strncmp(CHARS(arg),"0x",2)) {
	      
	      sscanf(CHARS(arg)+2,"%x",&i);
	      SET_COLOR(*bg_ptr,i);
	      
	    } else {
	      RefString shared=findRefString(all_buttons,n_all,arg);
	      if (shared)
		SET_PIXMAP(*bg_ptr,shared);
	      else
		SET_PIXMAP(*bg_ptr,arg);
	    }
	    
	    FREE_REF_STRING(arg);
	  }
	}
	break;


      case ICONPATH: /**** continue here *****/
	checkMem();
	if CHECK_COND(state==IN_COMMON,"iconpath") {
	  if (checkArg(buffer,&col,line,"Path")) {
	  
	    int j;
	    checkMem();
	    assert((npaths==0)==(icon_paths==NULL),
		   "Inconsistent data");
	    arg=readChars(buffer,&col,0);
	    assert(arg!=NULL,"String read is null!");
	    arg=checkHomePath(arg);
	    for(j=0;j<npaths;j++)
	      if (!strcmp(icon_paths[j],arg))
		break;
	    if (j==npaths) { /* new path */
	      npaths++;
	      trace( icon_paths=(char**)safeAllocPtr(ralloc(icon_paths,npaths*sizeof(char*))) );
	      icon_paths[npaths-1]=arg;
	      j=strlen(arg);
	      if (j>longest_path_size)
		longest_path_size=j;
	    } else {
	      fprintf(stderr,debug(YELLOW)
		      "Icon path %s was already defined in position %d\n"
		      debug(BLACK),arg,j+1);
	      disalloc(arg);
	    }
	  }
	}
	break;


      case BIND:
	checkMem();
	if CHECK_COND(state==IN_COMMON,"bind") {
	  if (checkArg(buffer,&col,line,"Key binding")) {
	    
	    checkMem();
	    /* arg is used as simple char* here */
	    arg=buffer+col;
	    col+=jumpWordNull(arg)-arg;
	    if (checkArg(buffer,&col,line,"Command")) {
	      i=findCode(arg);
	      if (i) {
		KeyBinding command=(KeyBinding)readChars(buffer,&col,sizeof(int));
		KEYCODE(command)=i;
		nkeybindings++;
		trace(keybindings=(KeyBinding*)safeAllocPtr(ralloc(keybindings,nkeybindings*sizeof(KeyBinding))));
		keybindings[nkeybindings-1]=command;
		debugPrint(("bind>"GREEN"%s"BLACK"(0x%x)="GREEN"%s"BLACK"\n",
			    arg,KEYCODE(command),KEYCOMMAND(command)));
	      } else {
		fprintf(stderr,
			debug(YELLOW)
			"Key symbol "debug(RED)"%s"debug(YELLOW)" not found!\n"
			debug(BLACK),arg);
		while(buffer[col]!='\n' && buffer[col]!=0) col++;
	      }
	    }
	  }
	}
	break;


      case INIT:
	checkMem();
	if CHECK_COND(state==IN_COMMON || state==IN_BUTTON,"init") {
	  if (checkArg(buffer,&col,line,"init command")) {

	    checkMem();
	    exec_command(buffer+col,NULL);
	    if (strncmp(buffer+col,"page",4)) {
	      nb_init++;
	      trace(init_commands=(InitCommand*)safeAllocPtr(ralloc(init_commands,nb_init*sizeof(InitCommand))));
	      init_commands[nb_init-1].init_command=safeAllocPtr(strdup(buffer+col));
	      addMemoryBlock(init_commands[nb_init-1].init_command,
			     strlen(init_commands[nb_init-1].init_command)+1);
	      init_commands[nb_init-1].button_id=(state==IN_COMMON ? -1 : n_all-1);
	    }
	    while(buffer[col]) col++;
	  }
	}
	break;


      case ICON:
	checkMem();
	if CHECK_COND(state==IN_COMMON || state==IN_BUTTON,"icon") {
	  if (checkArg(buffer,&col,line,"Icon path")) {

	    RefString shared;
	    
	    checkMem();
	    if (BUT(thebutton).icon)
	      FREE_REF_STRING(BUT(thebutton).icon);
	    arg=readString(buffer,&col);
	    shared=findRefString(all_buttons,n_all,arg);
	    assert(thebutton!=NULL,"Current button is null!");
	    if (shared) {
	      SET_REF_STRING(BUT(thebutton).icon,shared);
	      FREE_REF_STRING(arg);
	    } else
	      BUT(thebutton).icon=arg;
	  }
	}
	break;

      case X:
	caseX=1;
	arg="x";
	goto readXY;

      case Y:
	caseX=0;
	arg="y";
      readXY:
	checkMem();
	if CHECK_COND(state==IN_BUTTON || state==IN_COMMON ||
		      state==IN_DCLOCK || state==IN_MAILCHECK,arg) {
	  if (readIntArg(buffer,&col,line,&i)) {
	    assert(thebutton!=NULL,"Current button is null!");
	    if (caseX) thebutton->x=default_button.x=i;
	    else thebutton->y=default_button.y=i;
	  }
	}
	break;

      case XPLUS:
	caseX=1;
	arg="x+";
	goto readXYplus;

      case YPLUS:
	caseX=0;
	arg="y+";
      readXYplus:
	i=1; /* sign */
      readXYminus:
	checkMem();
	if CHECK_COND(state==IN_BUTTON || state==IN_COMMON ||
		      state==IN_DCLOCK || state==IN_MAILCHECK,arg) {
	  if (checkArg(buffer,&col,line,"Integer or \"width\" or \"height\"")) {

	    int savecol=col;
	    
	    checkMem();
	    assert(thebutton!=NULL,"Current button is null!");
	    switch(readToken(buffer,&col)) {
	    case WIDTH:
	      i*=thebutton->width;
	      break;
	      
	    case HEIGHT:
	      i*=thebutton->height;
	      break;
	      
	    default:
	      i*=readInt(buffer,&savecol);
	    }
	    if (caseX) thebutton->x=(default_button.x+=i);
	    else thebutton->y=(default_button.y+=i);
	  }
	}
	break;
	  
      case XMINUS:
	caseX=1;
	arg="x-";
	goto readYminus;

      case YMINUS:
	caseX=0;
	arg="y-";
      readYminus:
	i=-1;
	goto readXYminus;


      case TOOLTIP_FONT:
	checkMem();
	if CHECK_COND(state==IN_COMMON || state==IN_BUTTON ||
		      state==IN_DCLOCK || state==IN_MAILCHECK,
		      "tooltip_font") {
	  if (readIntArg(buffer,&col,line,&i)) {
	    if (checkArg(buffer,&col,line,"Font name")) {
	      trace(arg=readChars(buffer,&col,0));
	      trace(thebutton->tooltip_font=loadFont(arg,i));
	      disalloc(arg);
	    }
	  }
	}
	break;


      case FONT:
	checkMem();
	if (state==IN_DCLOCK_LABEL) {
	  if (IS_PERIODIC(thebutton) && clockData && clockData->label) {
	    assert(clockData->n>0,"In a label and no data!??");
	    if (readIntArg(buffer,&col,line,&i)) {
	      if (checkArg(buffer,&col,line,"Font name")) {
		trace(arg=readChars(buffer,&col,0));
		trace(clockData->label[clockData->n-1].font=loadFont(arg,i));
		disalloc(arg);
	      }
	    }
	  } else {
	    fprintf(stderr,debug(YELLOW)
		    "Parse error at line %d, column %d!"debug(BLACK)"\n",
		    line,col);
	    state=NOWHERE;
	  }
	} else if (state==IN_MAILCHECK) {
	  if (IS_PERIODIC(thebutton)) {
	    PopData *pop=(PopData*)PERIODIC(thebutton).data;
	    if (pop) {
	      if (readIntArg(buffer,&col,line,&i)) {
		if (checkArg(buffer,&col,line,"Font name")) {
		  trace(arg=readChars(buffer,&col,0));
		  trace(pop->font=loadFont(arg,i));
		  disalloc(arg);
		}
	      }
	      break;
	    }
	  }
	  fprintf(stderr,debug(YELLOW)
		  "Parse error at line %d, column %d!"debug(BLACK)"\n",
		  line,col);
	  state=NOWHERE; 
	} else
	  fprintf(stderr,debug(YELLOW)"Unexpected \"font\" at line %d, column %d\n"
		  debug(BLACK),line,col+1);
	break;


      case FORMAT:
	checkMem();
	if CHECK_COND(state==IN_DCLOCK_LABEL,"format") {
	  if (IS_PERIODIC(thebutton) && clockData && clockData->label) {
	    assert(clockData->n>0,"In a label and no data!??");
	    if (checkArg(buffer,&col,line,"Clock label format")) {
	      trace(arg=readChars(buffer,&col,0));
	      if (clockData->label[clockData->n-1].format)
		disalloc(clockData->label[clockData->n-1].format);
	      trace(clockData->label[clockData->n-1].format=arg);
	    }
	  } else {
	    fprintf(stderr,debug(YELLOW)
		    "Parse error at line %d, column %d!"debug(BLACK)"\n",
		    line,col);
	    state=NOWHERE;
	  }
	}
	break;

      case TOOLTIP_ICON:
	checkMem();
	if CHECK_COND(state==IN_COMMON,"tooltip_icon") {
	  if (readIntArg(buffer,&col,line,&i)) {
	    if (checkArg(buffer,&col,line,"Icon name")) {
	      int old_nb=nb_tooltip_icons;
	      if (i<1) i=1;
	      if (i>nb_tooltip_icons) {
		nb_tooltip_icons=i;
		trace(tooltip_icons=(ImlibImage**)safeAllocPtr(ralloc(tooltip_icons,nb_tooltip_icons*sizeof(char*))));
		for(;old_nb<i;old_nb++)
		  tooltip_icons[old_nb]=NULL;
	      }
	      arg=readChars(buffer,&col,0);
	      if (tooltip_icons[i-1])
		Imlib_kill_image(im_data,tooltip_icons[i-1]);
	      tooltip_icons[i-1]=getImage(im_data,arg,-1,-1);
	      debug(if (tooltip_icons[i-1])
		    printf(GREEN"Tooltip icon %d: %s OK!"BLACK"\n",
			   i,arg);
		    else
		    printf(YELLOW"Tooltip icon %d: %s NOT OK!"BLACK"\n",
			   i,arg));
	      disalloc(arg);
	    }
	  }
	}
	break;


      case TOOLTIP:
	caseX=1; /* eXtra one : odd */
	arg="tooltip";
	goto readAction;

      case ACTION:
	caseX=0;
	arg="action";
	/* PREVIOUS_PAGE NEXT_PAGE EXEC EXIT CONFIGURE ABOUT : dynamically */
      readAction:
	checkMem();
	if CHECK_COND(state==IN_COMMON || state==IN_BUTTON,arg) {
	  if (readIntArg(buffer,&col,line,&i)) {
	    if (i<1) {
	      fprintf(stderr,debug(YELLOW)
		      "An action/tooltip is coupled to a mouse button,"
		      " starting at 1 ; 1 taken instead of %d\n"
		      debug(BLACK),i);
	      i=1;
	    }
	    if (i>16) {
	      fprintf(stderr,debug(YELLOW)
		      "An action/tooltip is coupled to a mouse button,"
		      " you do not have %d buttons! 1 taken.\n"
		      debug(BLACK),i);
	      i=1;
	    }
	    if (checkArg(buffer,&col,line,"Command")) {
	      RefString shared;

	      assert(thebutton!=NULL,"Current button is null!");
	      if (i>NACTIONS(thebutton)) {
		int j;

		debugPrint(("Old nactions=%d ; new=%d\n",NACTIONS(thebutton),i));
		trace(BUT(thebutton).action=
		  (RefString*)safeAllocPtr(ralloc(BUT(thebutton).action,
					       i*2*sizeof(RefString))) );
		for(j=2*NACTIONS(thebutton);j<2*i;j++)
		  BUT(thebutton).action[j]=NULL; /* reset the new ones */
		NACTIONS(thebutton)=i;
	      }

	      arg=(RefString)readChars(buffer,&col,sizeof(int));
	      REF(arg)=1;
	      shared=findRefString(all_buttons,n_all,arg);
	      if (shared)
		SET_REF_STRING(BUT(thebutton).action[2*(i-1)+caseX],shared);
	      else
		SET_REF_STRING(BUT(thebutton).action[2*(i-1)+caseX],arg);
	      FREE_REF_STRING(arg);
	    }
	  }
	}
	break;


      default:
	checkMem();
	/* try all the options */
	for(i=0;i<NB_OPT;i++)
	  if (buttonRC[i].parseCode==token) {
	    if CHECK_COND(state==IN_BUTTON ||
			  state==IN_COMMON ||
			  state==IN_DCLOCK ||
			  state==IN_MAILCHECK,
			  buttonRC[i].name) {
	      assert(thebutton!=NULL,"Current button is null!");
	      trace( readIntArg(buffer,&col,line,&caseX) );
	      if (i==I_PAGE) caseX--;
	      if (i!=I_BELOW && caseX<0) caseX=0;
	      thebutton->gadget.any.opt[i]=caseX;
	      break;
	    }
	  }
	if (i==NB_OPT) {
	  fprintf(stderr,
		  debug(YELLOW)
		  "Unexpected token at line %d, column %d, skipping...\n"
		  debug(BLACK),line,col+1);
	}
      }
    }
  }
  fclose(f);
  freeButtonContent(&default_button);

  npages=0;
  /* sort the buttons per page */
  for(i=0;i<n_all;i++) {
    int j;
    if (PAGEOF(all_buttons+i)>npages)
      npages=PAGEOF(all_buttons+i);
    for(j=i+1;j<n_all;j++)
      if (PAGEOF(all_buttons+i)>PAGEOF(all_buttons+j)) {
	int k;
	default_button=all_buttons[i];
	all_buttons[i]=all_buttons[j];
	all_buttons[j]=default_button;
	/* and correct references for the init commands */
	for(k=0;k<nb_init;k++)
	  if (init_commands[k].button_id==i)
	    init_commands[k].button_id=j;
	  else if (init_commands[k].button_id==j)
	    init_commands[k].button_id=i;
      }
  }
  npages++;

  debugPrint(("Defined "GREEN"%d pages"BLACK"\n",npages));
  trace(pages=(ButtonPage*)safeAllocPtr(alloc(npages*sizeof(ButtonPage))));

  line=0; /* reuse it as index for the current page */
  /* put the buttons in the right page */
  for(i=0;i<npages;i++) {
    int j;
    /* count how many buttons */
    for(j=line;j<n_all && PAGEOF(all_buttons+j)==i;j++)
      debugPrint(("%d:%d: %s\n",j,PAGEOF(all_buttons+j),
		  (BUT(all_buttons+j).icon?CHARS(BUT(all_buttons+j).icon):"null")));
    /* write information */
    pages[i].nbuttons=j-line;
    pages[i].buttons=all_buttons+line;
    /* jump these buttons */
    line=j;
    debugPrint(("%d buttons in page %d\n",pages[i].nbuttons,i));
  }

  printMemory(); 
}


/* save the buttons
 */
void saveButtons() {
  FILE *f;
  pushMe();
  assert(configFile!=NULL,"Config file is not initialized!");
  f=fopen(configFile,"w");
  if (!f) {
    fprintf(stderr,"Cannot save buttons!\n");
    return;
  }
  trace( printButtons(f) );
  fclose(f);
  chmod(configFile,S_IRUSR|S_IWUSR);
}


void print_background(FILE *out,char *name,Background bg) {
  if (IS_TRANSPARENT(bg))
    fprintf(out,"  %s transparent\n",name);
  else if (IS_PIXMAP(bg))
    fprintf(out,"  %s %s\n",name,CHARS(GET_PIXMAP(bg)));
  else
    fprintf(out,"  %s 0x%x\n",name,GET_COLOR(bg));
}



/* print the buttons to a stream
 */
void printButtons(FILE *out) {
  RefString action[12]={0,0,0,0,0,0,0,0,0,0,0,0};
  GadgetUnion common;
  int i,n_all,width,height,lastpage=0,x,y,tooltip_bw=0,tooltip_bh=0;
  int colors[6];
  ButtonPage *page;
  Gadget *button;
  Background bg,tooltip_bg;
  FontInfo *tooltip_font=NULL;

  pushArgs(8,(BUF,"%s",PTR2STR(out)));
  assert(out!=NULL,"Output stream is null!");
  assert((npages==0)==(pages==NULL),"Inconsistent data");
  assert((npaths==0)==(icon_paths==NULL),"Inconsistent data");
  assert(pages==NULL || pages->buttons,"Inconsistent data");
  assert((extGadgets==NULL)==(nb_extGadgets==0),"Inconsistent data");

  if (pages && pages->buttons) {

    for(i=0;i<4;i++)
      common.raw.opt[i]=0xffffffff;
    width=pages->buttons->width;
    height=pages->buttons->height;
    bg=pages->buttons->bg;
    tooltip_bg=pages->buttons->tooltip_bg;
    tooltip_bw=pages->buttons->tooltip_bw;
    tooltip_bh=pages->buttons->tooltip_bh;
    memcpy(colors,pages->buttons->gradColor,6*sizeof(int));
    x=y=0;
    tooltip_font=NULL;
    
    /* look at the buttons first */
    for(i=0,page=pages,n_all=0;i<npages;i++,page++) {
      int n,j;
      n=page->nbuttons;
      n_all+=n;
      for(j=0;j<n;j++) {
	int k,na;
	button=page->buttons+j;
	assert(IS_BUTTON(button),"Button type expected");
	for(k=0;k<4;k++)
	  common.raw.opt[k]&=button->gadget.raw.opt[k];
	for(k=0,na=2*NACTIONS(button);k<na && k<12;k++) {
	  if (action[k]==NULL)
	    action[k]=BUT(button).action[k]; /* initialize */
	  else if (action[k]!=BUT(button).action[k])
	    action[k]=(RefString)1; /* invalidate */
	}
	for(;k<12;k++)
	  action[k]=(RefString)1; /* invalidate rest */
	if (tooltip_font==NULL)
	  tooltip_font=button->tooltip_font;
      }
    }      
    
    assert(n_all>0,"# buttons inconsistent with general pointer");
    if (n_all==0) /* just in case */
      return;
    
    button=pages->buttons; /* general pointer */
    
    fprintf(out,
	    "# Automatically generated file, you can modify it but comments will be lost\n"
	    "\n"
	    "common {\n"
	    "\n"
	    "# To restart on the same page\n"
	    "  init page %d\n"
	    "\n"
	    "# Percentage of darkness when a button is pressed\n"
	    "  darker %d\n"
	    "\n"
	    "# Percentage of light when a button is focused\n"
	    "  lighter %d\n"
	    "\n"
	    "# Tooltip waiting time in seconds\n"
	    "  tooltip_wait %g\n"
	    "\n"
	    "# Tooltip font colors\n"
	    ,current_page+1
	    ,(100*(255-darker))/255
	    ,(100*(lighter-255))/255
	    ,tooltip_period);
    
    for(i=0;i<6;i++)
      fprintf(out,"  %s %d\n",color_names[i],colors[i]);
    
    if (tooltip_font)
      fprintf(out,"\n# Tooltip font\n  tooltip_font %c %s\n",
	      tooltip_font->name[0],
	      tooltip_font->name+1);
    
    
    for(i=0;i<NB_OPT;i++) {
      assert((buttonRC[i].parseCode==0)==(buttonRC[i].comment==0),"Inconsistent data");
      assert(buttonRC[i].name!=NULL,"Bad data");
      assert(buttonRC[i].parseCode!=0 ||
	     (buttonRC[i].name!=NULL && strlen(buttonRC[i].name)==0),
	     "parseCode==0 => len(name)==0");/* This ^^^^^^^ is to avoid seg fault if pb */
      
      if (buttonRC[i].comment) {
	int val=common.any.opt[i];
	fprintf(out,"\n# %s\n  %s %d\n",
		buttonRC[i].comment,
		buttonRC[i].name,
		i==I_PAGE?(lastpage=val)+1:val);
      }
    }
    
    if (width>0)
      fprintf(out,"\n"
	      "# Common width of the icons\n"
	      "  width %d\n",width);
    
    if (height>0)
      fprintf(out,"\n"
	      "# Common height of the icons\n"
	      "  height %d\n",height);
    
    if (tooltip_bw>0)
      fprintf(out,"\n"
	      "# Common width of the tooltip border\n"
	      "  tooltip_border_width %d\n",tooltip_bw);
    
    if (tooltip_bh>0)
      fprintf(out,"\n"
	      "# Common height of the tooltip border\n"
	      "  tooltip_border_height %d\n",tooltip_bh);
    
    assert(npaths>=0,"Corrupted npath");
    assert((npaths==0)==(icon_paths==NULL),"Inconsistent data");
    if (npaths>0) {
      fprintf(out,"\n"
	      "# Path for the icons\n");
      for(i=0;i<npaths;i++) {
	assert(icon_paths[i]!=NULL,"Null path!");
	fprintf(out,"  iconpath %s\n",icon_paths[i]);
      }
    }
    
    fprintf(out,"\n# Background: transparent/none 0xBBGGRR or a pixmap path\n");
    print_background(out,"background",bg);
    
    fprintf(out,"\n# Tooltip background: transparent/none 0xBBGGRR or a pixmap path\n");
    print_background(out,"tooltip_background",tooltip_bg);
    
    assert((nb_tooltip_icons==0)==(tooltip_icons==NULL),"Inconsistent data");
    fprintf(out,"\n# Tooltip icons\n");
    for(i=0;i<nb_tooltip_icons;i++)
      if (tooltip_icons[i]) {
	if (tooltip_icons[i]->filename)
	  fprintf(out,"  tooltip_icon %d %s\n",i+1,tooltip_icons[i]->filename);
      }
    
    assert(nkeybindings>=0,"Corrupted nkeybindings");
    assert((nkeybindings==0)==(keybindings==NULL),"Inconsistent data");
    if (nkeybindings>0) {
      fprintf(out,"\n"
	      "# Key bindings: see /usr/include/X11/keysymdef.h for the symbols\n"
	      "# They are the letter definitions without XK_ for convenience\n");
      for(i=0;i<nkeybindings;i++) {
	char *symbol;
	assert(keybindings[i]!=NULL,"Null key binding");
	trace(symbol=findSymbol(KEYCODE(keybindings[i])));
	if (symbol) {
	  fprintf(out,"  bind %s %s\n",symbol,KEYCOMMAND(keybindings[i]));
	  disalloc(symbol);
	} else
	  fprintf(stderr,
		  "Panic: could not find symbol(code=0x%x) again! Key lost.\n",
		  KEYCODE(keybindings[i]));
      }
    }
    
    if (nb_init>0) {
      fprintf(out,"\n"
	      "# General initializations\n");
      for(i=0;i<nb_init;i++)
	if (init_commands[i].button_id<0) /* general marker */
	  fprintf(out,"  init %s",init_commands[i].init_command);
    }
    
    fprintf(out,"\n");
    for(i=0;i<6;i++) {
      if (action[2*i]>(RefString)1)
	fprintf(out,"  action %d %s\n",i+1,CHARS(action[2*i]));
      if (action[2*i+1]>(RefString)1)
	fprintf(out,"  tooltip %d %s\n",i+1,CHARS(action[2*i+1]));
    }
    fprintf(out,"}\n");
    
    for(i=0;i<n_all;i++,button++) {
      int j,na;
      
      if (lastpage!=PAGEOF(button))
	fprintf(out,"\ncommon { # new line now, compulsory\n  page %d\n}\n",
		(lastpage=PAGEOF(button))+1);
      
      fprintf(out,"\nbutton {\n");
      
      if (width<=0)
	fprintf(out,"  x %d\n  width %d\n",button->x,button->width);
      else {
	if (i==0 || abs(button->x-x)!=width) {
	  if (i==0 || x!=button->x)
	    fprintf(out,"  x %d\n",button->x);
	} else if (x<button->x)
	  fprintf(out,"  x+ width\n");
	else if (x>button->x)
	  fprintf(out,"  x- width\n");
	
	x=button->x;
	
	if (width!=button->width)
	  fprintf(out,"  width %d\n",button->width);
      }
      
      if (height<=0)
	fprintf(out,"  y %d\n  height %d\n",button->y,button->width);
      else {
	if (i==0 || abs(button->y-y)!=height) {
	  if (i==0 || y!=button->y)
	    fprintf(out,"  y %d\n",button->y);
	} else if (y<button->y)
	  fprintf(out,"  y+ height\n");
	else if (y>button->y)
	  fprintf(out,"  y- height\n");
	
	y=button->y;
	
	if (height!=button->height)
	  fprintf(out,"  height %d\n",button->height);;
      }
      
      for(j=0;j<NB_OPT;j++)
	if (j!=I_PAGE && buttonRC[j].comment && common.any.opt[j]!=BUT(button).opt[j]) {
	  assert((buttonRC[j].parseCode==0)==(buttonRC[j].comment==0),"Inconsistent data");
	  assert(buttonRC[j].name!=NULL,"Bad data");
	  assert(buttonRC[j].parseCode!=0 || (buttonRC[j].name!=NULL && strlen(buttonRC[j].name)==0),
		 "parseCode==0 => len(name)==0");/* This ^^^^^^^ is to avoid seg fault if pb */
	  
	  fprintf(out,"\n# %s\n  %s %d\n",
		  buttonRC[j].comment,
		  buttonRC[j].name,
		  BUT(button).opt[j]);
	}
      
      if (BUT(button).icon)
	fprintf(out,"  icon %s\n",CHARS(BUT(button).icon));
      
      if (button->bg.colorCode!=bg.colorCode)
	print_background(out,"background",button->bg);
      
      for(j=0;j<nb_init;j++)
	if (i==init_commands[j].button_id)
	  fprintf(out,"  init %s",init_commands[j].init_command);
      
      assert((NACTIONS(button)==0)==(BUT(button).action==NULL),"Inconsistent data");
      na=NACTIONS(button);
      
      for(j=0;j<na;j++) {
	if (BUT(button).action[2*j] &&
	    (j>=6 || BUT(button).action[2*j]!=action[2*j]))
	  fprintf(out,"  action %d %s\n",j+1,CHARS(BUT(button).action[2*j]));
	if (BUT(button).action[2*j+1] &&
	    (j>=6 || BUT(button).action[2*j+1]!=action[2*j+1]))
	  fprintf(out,"  tooltip %d %s\n",j+1,CHARS(BUT(button).action[2*j+1]));
      }
      
      if (button->tooltip_font &&
	  button->tooltip_font!=tooltip_font)
	fprintf(out,"  tooltip_font %c %s\n",
		button->tooltip_font->name[0],
		button->tooltip_font->name+1);
      
      if (button->tooltip_bg.colorCode!=tooltip_bg.colorCode)
	print_background(out,"tooltip_background",button->tooltip_bg);
      
      if (button->tooltip_bw!=tooltip_bw && button->tooltip_bw>0)
	fprintf(out,"  tooltip_border_width %d\n",button->tooltip_bw);
      
      if (button->tooltip_bh!=tooltip_bh && button->tooltip_bh>0)
	fprintf(out,"  tooltip_border_height %d\n",button->tooltip_bh);
      
      for(j=0;j<6;j++)
	if (colors[j]!=button->gradColor[j]) {
	  for(j=0;j<6;j++)
	    fprintf(out,"  %s %d\n",color_names[j],button->gradColor[j]);
	  break;
	}
      
      fprintf(out,"}\n");  
    }
  }

  button=extGadgets;
  n_all=nb_extGadgets;
  for(i=0;i<n_all;i++,button++) {
    int j;

    if (IS_PERIODIC(button)) {
      PERIODIC(button).saveGadget(out,button); /* specific */
      fprintf(out,"  update %g\n",
	      PERIODIC(button).update_period);
    }

    /* common now */
    
    fprintf(out,
	    "  x %d\n"
	    "  y %d\n"
	    "  width %d\n"
	    "  height %d\n",
	    button->x,button->y,
	    button->width,button->height);

    for(j=0;j<COMMON_NB_OPT;j++)
      if (buttonRC[j].comment && common.any.opt[j]!=ANY(button).opt[j]) {
	assert((buttonRC[j].parseCode==0)==(buttonRC[j].comment==0),"Inconsistent data");
	assert(buttonRC[j].name!=NULL,"Bad data");
	assert(buttonRC[j].parseCode!=0 || (buttonRC[j].name!=NULL && strlen(buttonRC[j].name)==0),
	       "parseCode==0 => len(name)==0");/* This ^^^^^^^ is to avoid seg fault if pb */
	  
	fprintf(out,"\n# %s\n  %s %d\n",
		buttonRC[j].comment,
		buttonRC[j].name,
		ANY(button).opt[j]);
      }
    
    if (button->bg.colorCode!=bg.colorCode)
      print_background(out,"background",button->bg);
    
    if (button->tooltip_font &&
	button->tooltip_font!=tooltip_font)
      fprintf(out,"  tooltip_font %c %s\n",
	      button->tooltip_font->name[0],
	      button->tooltip_font->name+1);
    
    if (button->tooltip_bg.colorCode!=tooltip_bg.colorCode)
      print_background(out,"tooltip_background",button->tooltip_bg);
    
    if (button->tooltip_bw!=tooltip_bw && button->tooltip_bw>0)
      fprintf(out,"  tooltip_border_width %d\n",button->tooltip_bw);
    
    if (button->tooltip_bh!=tooltip_bh && button->tooltip_bh>0)
      fprintf(out,"  tooltip_border_height %d\n",button->tooltip_bh);
    
    for(j=0;j<6;j++)
      if (colors[j]!=button->gradColor[j]) {
	for(j=0;j<6;j++)
	  fprintf(out,"  %s %d\n",color_names[j],button->gradColor[j]);
	break;
      }
    
    fprintf(out,"}\n");
  }
  
  fflush(out);
}



/* initialize the configuration file name
 */
void initConfigFile(char *otherConfig) {
  char *home=getenv("HOME");
  char *config=(otherConfig?otherConfig:CONFIGFILE);
  int len;

  pushMe();
  if (!home) {
    fprintf(stderr,"HOME variable not setup, cannot find home directory, aborting!\n");
    exit(1);
  }
  len=strlen(home)+sizeof(CONFIGPATH)+strlen(config)+2;
  if (configFile) free(configFile);
  trace(configFile=(char*)safeAllocPtr(malloc(len))); /* not alloc because this is apart */
  sprintf(configFile,"%s/"CONFIGPATH,home);
  checkDirectory(configFile);
  sprintf(configFile,"%s/"CONFIGPATH"/%s",home,config);
  debugPrint(("Configuration file set to: %s\n",configFile));
}


/* code and decode passwd: depends on
 * the compiled binary, the user running the program and
 * the password saved in the file (in principle not visible)
 */

void decode_passwd(char *key,char *passwd) {
  char salt[3]={SALT_0,SALT_1,0}; /* to avoid the string in the data */
  char *crypted_key=crypt(key,salt);
  int uid=getuid();
  pushArgs(STRSIZE(key)+STRSIZE(passwd)+8,(BUF,"%s,%s",key,passwd));
  /* debugPrint((GREEN"Decode %s with %s\n"BLACK,passwd,key)); */
  if (*passwd) {
    int lenk=strlen(crypted_key)+strlen(CRYPT_KEY);
    char crypted[lenk+1];
    int lenp=strlen(passwd);
    int i,j;
    trace(strcpy(crypted,crypted_key));
    trace(strcat(crypted,CRYPT_KEY));
    for(i=lenk-1,j=(lenk-1)%lenp;i>=0;i--) {
      if (j<0)
	j=lenp-1;
      trace(passwd[j--]^=(uid*(crypted[i]^CRYPT_MASK))&0x5f);
    }
    assert(j==-1,"Beginning not reached");
    assert(i==j,"Wrong counters");
  }
  /* debugPrint(("Result: %s\n",passwd)); */
}

void code_passwd(char *key,char *passwd) {
  char salt[3]={SALT_0,SALT_1,0}; /* to avoid the string in the data */
  char *crypted_key=crypt(key,salt);
  char *pass=passwd;
  int uid=getuid();
  pushArgs(STRSIZE(key)+STRSIZE(passwd)+8,(BUF,"%s,%s",key,passwd));
  /* debugPrint((GREEN"Code %s with %s\n"BLACK,passwd,key)); */
  if (*passwd) {
    char crypted[strlen(crypted_key)+strlen(CRYPT_KEY)+1];
    int i=0;
    trace(strcpy(crypted,crypted_key));
    trace(strcat(crypted,CRYPT_KEY));
    while(crypted[i]) {
      trace(*pass^=(uid*(crypted[i]^CRYPT_MASK))&0x5f);
      i++;
      pass++;
      trace(if (!*pass)	pass=passwd);
    }
  }
  /* debugPrint(("Result: %s\n",passwd)); */
}

