#include <stdio.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "debug.h"
#include "utils.h"
#include "gadget.h"
#include "mailcheck.h"
#include "load_save.h"
#include "Xcore.h"
#include "pixmaps.h"

#define TIMEOUT 90.0    /* 1 minute 30 sec */
#define ERRORWAIT 600.0 /* 10 minutes */

#define BUFSIZE 1024
#define LINESIZE 128

static int open_connection(char *server) {
  struct hostent *host;
  struct sockaddr_in sa_in;
  int sock_in;
  
  debugPrint(("gethost %s\n",server));
  host=gethostbyname(server);
  if (host==NULL) {
    perror("gethostbyname");
    return -1;
  }
  memcpy (&sa_in.sin_addr, host->h_addr, sizeof (struct in_addr));

  debugPrint(("creating socket\n"));
  sock_in=socket(AF_INET,SOCK_STREAM,0);
  if (sock_in<0) {
    perror("socket");
    return -1;
  }

  debugPrint(("connecting\n"));
  sa_in.sin_family=AF_INET;
  sa_in.sin_port=htons(110);
  if (connect(sock_in,(struct sockaddr *)&sa_in,sizeof(sa_in))<0) {
    perror("connect");
    close(sock_in);
    return -1;
  }

  return sock_in;
}

static int read_info(PopInfo *pop_info,char *buffer) {
  debugPrint(("reading information\n"));
  memset(buffer,0,BUFSIZE);
  if (read(pop_info->fd,buffer,BUFSIZE)<=0 ||
      strncmp(buffer,"+OK",3)) {
    close(pop_info->fd);
    pop_info->fd=-1;
    pop_info->nb_msg=0;
    debugPrint(("Error in state %d: %s\n",pop_info->state,buffer));
    fprintf(stderr,"Pop server error: %s\n",buffer);
    pop_info->state=-1;
    return 1;
  }
  return 0;
}

static void get_pop(Gadget *gad,
		    double *next_time,fd_set *fdset,
		    PopInfo *pop_info,double now) {
  char buffer[BUFSIZE];
  char *p;

  debugPrint(("get_pop: state %d\n",pop_info->state));

  if (pop_info->state>0) {   /* was waiting for an answer */
    assert(pop_info->fd>=0,"Inconsistent state");
    if (1||FD_ISSET(pop_info->fd,fdset)) { /* answer has arrived */
      if (read_info(pop_info,buffer)) { /* if reading error */
	pop_info->state=-1;
	close(pop_info->fd);
	pop_info->fd=-1;
	pop_info->time=now+ERRORWAIT;
	return;
      }
    } else if (pop_info->time+TIMEOUT<=now) { /* time out */
      close(pop_info->fd);
      pop_info->fd=-1;
      pop_info->time=now+ERRORWAIT;
      pop_info->state=-1;
      fprintf(stderr,"Pop server timed out!\n");
      return;
    } else {              /* update next time, not called */
      if (*next_time>pop_info->time+TIMEOUT-now)
	*next_time=pop_info->time+TIMEOUT-now;
      return;
    }
  }
  
  switch(pop_info->state) {

  case -2: /* wait next update */
  case -1: /* previous error */
    if (pop_info->time>now)
      return;
    /* else retry */

  case 0: /* start */
    pop_info->fd=open_connection(pop_info->server);
    if (pop_info->fd<0) {
      pop_info->state=-1;
      pop_info->time=now+ERRORWAIT;
      return;
    }
    pop_info->state=0; /* if from error/next update */
    debugPrint(("Connection opened\n"));
    break;

  case 1: /* pop client information arrived */
    debugPrint(("Pop server: %s\n",buffer));
    snprintf(buffer,LINESIZE,"USER %s\r\n",pop_info->username);
    write(pop_info->fd,buffer,strlen(buffer));
    break;

  case 2: /* username answer */
    debugPrint(("Username: %s\n",buffer));
    snprintf(buffer,LINESIZE-2,"PASS %s",pop_info->passwd);
    trace(decode_passwd(pop_info->crypt_pass,buffer+5));
    strcat(buffer,"\r\n");
    write(pop_info->fd,buffer,strlen(buffer));
    break;

  case 3: /* password answer */
    debugPrint(("Password: %s\n",buffer));
    write(pop_info->fd,"STAT\n",5);
    break;

  case 4: /* stat answer */
    debugPrint(("Stat: %s\n",buffer));
    p=buffer+4;
    while(*p && *p!=' ') p++;
    *p=0;
    pop_info->nb_msg=atoi(buffer+4);
    write(pop_info->fd,"QUIT\n",5);
    break;

  case 5: /* quit answer */
    debugPrint(("Quit: %s\n",buffer));
    close(pop_info->fd);
    pop_info->fd=-1;
    pop_info->state=-2;
    pop_info->time=now+PERIODIC(gad).update_period;
    return;

  }
  
  FD_SET(pop_info->fd,fdset);
  pop_info->state++;
  pop_info->time=now;
  *next_time=0;

}

void addMailFD(Gadget *gad,int *nfd,fd_set *fdset) {
  printHere();
  if (gad) {
    PopData *pop=(PopData*)PERIODIC(gad).data;
    int i,n=pop->nb_pop;
    for(i=0;i<n;i++)
      if (pop->pop_info[i].fd>=0) {
	FD_SET(pop->pop_info[i].fd,fdset);
	if (*nfd<pop->pop_info[i].fd)
	  *nfd=pop->pop_info[i].fd;
      }
  }
}

int calledMailGadget(Gadget *gad,fd_set *fdset) {
  printHere();
  if (gad) {
    PopData *pop=(PopData*)PERIODIC(gad).data;
    int i,n=pop->nb_pop;
    for(i=0;i<n;i++)
      if (pop->pop_info[i].fd>=0 &&
	  FD_ISSET(pop->pop_info[i].fd,fdset))
	return 1;
  }
  return 0;
}

void saveMailGadget(FILE* out,Gadget *gad) {
  PopData *pop;
  printHere();
  pop=(PopData*)PERIODIC(gad).data;
  fprintf(out,"\nmail_check {\n");
  if (pop) {
    int i,n;
    n=pop->nb_pop;
    for(i=0;i<n;i++)
      fprintf(out,"  coded_passwd %s %s %s %s\n",
	      pop->pop_info[i].crypt_pass,
	      pop->pop_info[i].username,
	      pop->pop_info[i].server,
	      pop->pop_info[i].passwd);
    for(i=0;i<8;i++)
      fprintf(out,"  %s %d\n",labelnames[i],pop->color[i]);
    if (pop->font)
      fprintf(out,"  font %c %s\n",
	      pop->font->name[0],
	      pop->font->name+1);
    print_background(out,"new_mail_bg",pop->new_bg);
    print_background(out,"error_mail_bg",pop->error_bg);
    if (pop->command)
      fprintf(out,"  mail %s\n",pop->command);
  }
  debugPrint(("\n"));
}

static void setup_pixmap(ImlibImage **orig,
			 Background bg,
			 ImlibImage **normal,
			 Gadget *gad) {
  if (!*orig) {
    trace(*orig=IS_PIXMAP(bg)?getImage(im_data,CHARS(GET_PIXMAP(bg)),
				       gad->width,
				       gad->height) : NULL);
    if (!*orig) {
      if (IS_TRANSPARENT(bg)) {
	ImlibColor imcol={1,1,1};
	trace(*orig=getSolidImage(gad->width,gad->height,
				  im_data,0x010101));
	Imlib_set_image_shape(im_data,*orig,&imcol);
      } else {
	trace( *orig=getSolidImage(gad->width,gad->height,
				   im_data,IS_COLOR(bg)
				   ? GET_COLOR(bg)
				   : 0x9000cc ));
      }
    }
  }
  if (*orig) {
    if (!*normal)
      trace(*normal=Imlib_clone_image(im_data,*orig));
    else {
      trace(memcpy((*normal)->rgb_data,
		   (*orig)->rgb_data,
		   3*gad->width*gad->height));
      Imlib_changed_image(im_data,*normal);
    }
  }
}

double updateMailGadget(Gadget *gad,int *nfd,fd_set *fdset,int force) {
  double next_time=PERIODIC(gad).update_period;
  double now=get_time();
  Background bg={colorCode:0};
  int nb_msg=0;
  int bg_set=0;
  int nb_errors=0;
  int need_update=(PERIODIC(gad).im_normal==NULL);
  PopData *pop=(PopData*)PERIODIC(gad).data;
  assert(pop,"Null data");
  assert(gad,"Null gadget!");
  assert(gad==NULL || IS_PERIODIC(gad),"Bad type");
  printHere();
  if (!pop) return next_time;
  if (PERIODIC(gad).data) {
    int i,n;
    n=pop->nb_pop;
    for(i=0;i<n;i++) {
      if (force && pop->pop_info[i].state==-2)
	pop->pop_info[i].state=0;
      get_pop(gad,&next_time,fdset,pop->pop_info+i,now);
      nb_msg+=pop->pop_info[i].nb_msg;
      debugPrint(("%d messages on server %s\n",
		  pop->pop_info[i].nb_msg,
		  pop->pop_info[i].server));
      if (pop->pop_info[i].state==-1) {
	pop->pop_info[i].nb_msg=0;
	need_update|=1;
	nb_errors++;
      } else if (pop->pop_info[i].fd>*nfd)
	*nfd=pop->pop_info[i].fd;
      if (pop->pop_info[i].last_nb_msg!=pop->pop_info[i].nb_msg &&
	  pop->pop_info[i].state<0) {
	need_update|=1;
	if (pop->pop_info[i].last_nb_msg<pop->pop_info[i].nb_msg)
	  need_update|=2;
	pop->pop_info[i].last_nb_msg=pop->pop_info[i].nb_msg;
      }
    }
    debugPrint(("Total messages: %d\nErrors: %d\n",nb_msg,nb_errors));
    /* and choose the right icon based on that */
    bg=( nb_errors
	 ? pop->error_bg
	 : ( nb_msg
	     ? pop->new_bg
	     : gad->bg ));
    bg_set=1;
  }
  /* nothing found, take error pixmap */
  if (!bg_set) {
    nb_errors++;
    bg=pop->error_bg;
  }
  
  if (need_update) {
    if (need_update>1)
      exec_command(pop->command,gad);
    if (nb_errors) {
      setup_pixmap(&pop->error_mail,bg,&PERIODIC(gad).im_normal,gad);
      if (nb_msg)
	goto print_pop;
    } else if (nb_msg) {
      setup_pixmap(&pop->new_mail,bg,&PERIODIC(gad).im_normal,gad);
    print_pop:
      if (PERIODIC(gad).im_normal) {
	char buf[32];
	snprintf(buf,sizeof(buf),"Pop:%d",nb_msg);
	trace(printStringImage(pop->color[6],
			       pop->color[7],
			       PERIODIC(gad).im_normal,
			       buf,
			       pop->font,
			       pop->color));
	debugPrint(("Print: %s\n",buf));
      }
    } else {
      setup_pixmap(&PERIODIC(gad).im_orig,bg,&PERIODIC(gad).im_normal,gad);
    }

    if (PERIODIC(gad).im_normal) {
      Imlib_apply_image(im_data,PERIODIC(gad).im_normal,gad->win);
      XFlush(display);
    }
  }

  return next_time;
}


void freeMailGadget(Gadget *gad) {
  PopData *pop;
  printHere();
  pop=(PopData*)PERIODIC(gad).data;
  if (pop) {
    int i,n;
    assert((pop->nb_pop==0)==(pop->pop_info==NULL),"Inconsistent data");
    n=pop->nb_pop;
    for(i=0;i<n;i++) {
      FREE(pop->pop_info[i].server);
      FREE(pop->pop_info[i].passwd);
      FREE(pop->pop_info[i].username);
      FREE(pop->pop_info[i].crypt_pass);
      if (pop->pop_info[i].fd)
	close(pop->pop_info[i].fd);
    }
    if (pop->new_mail)
      Imlib_kill_image(im_data,pop->new_mail);
    if (pop->error_mail)
      Imlib_kill_image(im_data,pop->error_mail);
    if (IS_PIXMAP(pop->new_bg))
      FREE_REF_STRING(GET_PIXMAP(pop->new_bg));
    if (IS_PIXMAP(pop->error_bg))
      FREE_REF_STRING(GET_PIXMAP(pop->error_bg));
    FREE(pop->pop_info);
    FREE(pop->command);
    disalloc(pop);
    PERIODIC(gad).data=NULL;
  }
  debugPrint(("\n"));
}

char *tooltipMailGadget(Gadget *gad) {
  static char pop_buffer[512];
  printHere();
  debugPrint(("\n"));
  strcpy(pop_buffer,"Pop stat:\n");
  if (gad) {
    PopData *pop=(PopData*)PERIODIC(gad).data;
    int i,n=pop->nb_pop;
    for(i=0;i<n;i++) {
      int len=strlen(pop_buffer);
      if (pop->pop_info[i].state!=-1)
	snprintf(pop_buffer+len,sizeof(pop_buffer)-len,"%s:%d\n",
		 pop->pop_info[i].server,
		 pop->pop_info[i].nb_msg);
      else
	snprintf(pop_buffer+len,sizeof(pop_buffer)-len,"%s:error\n",
		 pop->pop_info[i].server);
    }
  }
  return pop_buffer;
}

void enteredMailGadget(Gadget *gad) {
  printHere();
  debugPrint(("\n"));
}

void leftMailGadget(Gadget *gad) {
  debugPrint(("\n"));
  printHere();
}


