/* pixmaps: functions to manipulate pixmaps
 * and RGB data. See Imlib2 to get a good explanation
 * on the used formulas.
 *
 * author: Alexandre David
 * license: GPL
 * e-mail: adavid@docs.uu.se
 * http://www.docs.uu.se/~adavid
 */

#include "debug.h"
#include "utils.h"
#include "gadget.h"
#include "pixmaps.h"

Display *display=0;
ImlibData *im_data=0;


/* p: pix to saturate
 * m: mask to use
 */
#define sat(p,m)\
 ((p) & ((m) - ((m) >> 8))) | (((p) & (m)) - (((p) & (m)) >> 8))


#define darkerChar(p)\
 (p) = (p) - ((p) >> 2)

#define lighterChar(p)\
 (p) = sat( ((unsigned int)(p)) << 1 , 0x100)


static void effectAt(unsigned char *ptr,int isLighter) {
  pushArgs(32,(BUF,"%s,%d",PTR2STR(ptr),isLighter));

  if (isLighter) {
    int i;
    for(i=0;i<3;i++)
      trace( lighterChar(ptr[i]) );
  } else {
    int i;
    for(i=0;i<3;i++)
      trace( darkerChar(ptr[i]) );
  }
}

/* apply effect lighter/darker on a button
 * skip 2 pixel border
 * width,height: wished dimension to modify
 * x: effect: 0-511 ; 0-255=darker ; 256-511=lighter
 */
void applyEffect(ImlibImage *im,int width,int height,int x,int border) {
  unsigned char *ptr;
  int data_w3,i;

  pushArgs(256,(BUF,"%s,%d,%d,%d,%d",
		PTR2STR(im),width,height,x,border));

  assert(im!=NULL,"Null image pointer");
  assert(width>=0,"Bad width value");
  assert(height>=0,"Bad height value");
  assert(x>=0,"Bad effect value");
  assert(x<512,"Bad effect value");
  assert(im->rgb_width>=0,"Bad image width");
  assert(im->rgb_height>=0,"Bad image height");
  assert(border>=0,"Bad border value");

  debugPrint(("applyEffect(im[%dx%d],width=%d,height=%d,x=%d,border=%d)\n",
	      im->rgb_width,im->rgb_height,width,height,x,border));

  width=3*(width-border);
  data_w3=im->rgb_width*3;
  if (width>=data_w3)
    width=data_w3;
  width-=3*border;

  height-=border;
  if (height>im->rgb_height)
    height=im->rgb_height;
  height-=border;

  ptr=im->rgb_data+border*data_w3+3*border; /* jump 2 lines,2 columns */

  for(;height>0;height--,ptr+=data_w3)
    for(i=0;i<width;i++) {
      unsigned int pix = (((int)ptr[i]) * x) >> 8;
      trace( ptr[i] = sat(pix,0x100) );
    }

  /* only a char at a time for big/little endian
   * compatibility reasons
   */
}


void paintBorder(ImlibImage *im,int width,int height,
		 char effect1,char effect2) {
  unsigned char *ptr;
  int data_w3,i,n,resth;
  
  pushArgs(256,(BUF,"%s,%d,%d,%d,%d",
		PTR2STR(im),width,height,effect1,effect2));

  assert(im!=NULL,"Null image pointer");
  assert(width>=0,"Bad width value");
  assert(height>=0,"Bad height value");
  assert(im->rgb_width>=0,"Bad image width");
  assert(im->rgb_height>=0,"Bad image height");

  debugPrint(("paintBorder(im[%dx%d],width=%d,height=%d,effect1=%d,effect2=%d)\n",
	      im->rgb_width,im->rgb_height,width,height,effect1,effect2));
  
  data_w3=3*im->rgb_width;
  width*=3;
  n=( width<=data_w3 ? width : data_w3);
  if (n==0) return;
  ptr=im->rgb_data;

  if (im->rgb_height>0) {
    /* first line */
    for(i=0;i<n;i+=3)
      trace( effectAt(ptr+i,effect1) );

    ptr+=data_w3; /* go to second line */
    
    if (im->rgb_height>1) {
      /* 1st col */
      trace( effectAt(ptr,effect1) );

      /* middle */
      for(i=3;i<width-3 && i<n;i+=3)
	trace( effectAt(ptr+i,effect2) );

      /* last col */
      if (i<n)
	trace( effectAt(ptr+i,~effect1) );

      ptr+=data_w3; /* go to 3rd line */
      resth=im->rgb_height-height+2;
      height=( height-3<=im->rgb_height-3 ? height-3 : im->rgb_height-3 );

      for(;height>0;height--,ptr+=data_w3) {
	trace( effectAt(ptr,effect1) );
	if (3<n) {
	  trace( effectAt(ptr+3,effect2) );
	  if (width-6<n) {
	    trace( effectAt(ptr+width-6,~effect2) );
	    if (width-3<n)
	      trace( effectAt(ptr+width-3,~effect1) );
	  }
	}
      }
	
      /* last-1 line */
      if (resth>0) {
	ptr-=data_w3;
	for(i=6;i<width-6 && i<n;i+=3)
	  trace( effectAt(ptr+i,~effect2) );
	
	/* last line */
	if (resth>1) {
	  ptr+=data_w3;
	  trace( effectAt(ptr,effect1) );
	  for(i=3;i<n;i+=3)
	    trace( effectAt(ptr+i,~effect1) );
	}
      }
    }
  }
  
}


ImlibImage* getSolidImage(int width,int height,
			  ImlibData *id,int rgbpix) {
  int i,n=width*height*3;
  char *data;
  ImlibImage *im;

  pushArgs(256,(BUF,"%d,%d,%s,%x",
		width,height,PTR2STR(id),rgbpix));
  trace(data=(char*)alloc(n));

  if (data==NULL) {
    fprintf(stderr,"Memory exhausted!\n");
    return NULL;
  }

  for(i=0;i<n;i+=3) {
    trace( data[i]=rgbpix );
    trace( data[i+1]=rgbpix>>8 );
    trace( data[i+2]=rgbpix>>16 );
  }

  trace( im=Imlib_create_image_from_data(id,data,NULL,width,height) );
  disalloc(data);

  return im;
}


/* take a snapshot of the screen ; or the root
 */
ImlibImage* snapshotImage(ImlibData* id,
			  int x,int y,int width,int height,
			  Window win) {
  Window root;

  pushArgs(256,(BUF,"id=%s,x=%d,y=%d,width=%d,height=%d,win=%x",
		PTR2STR(id),x,y,width,height,(int)win));

  assert(id!=NULL,"Null image data pointer");

  debugPrint(("snapshotImage(x=%d,y=%d,width=%d,height=%d,%s)\n",
	      x,y,width,height,win==0?"window":"root"));

  root=XDefaultRootWindow(display);

  if (win) {
    Window parent,*list;
    unsigned int n;
    parent=win;
    while(parent!=root) {
      XSetWindowBackgroundPixmap(display,parent,ParentRelative);
      XQueryTree(display,parent,&root,&parent,&list,&n);
      if (list) XFree(list);
    }
    XClearWindow(display,win);
    XFlush(display);
    XSync(display,False);
    x=0;
    y=0;
  } else
    win=root;

  return Imlib_create_image_from_drawable(id,win,None,x,y,width,height);
}


#define scale_image(id,im,width,height) \
 ( (width!=im->rgb_width || height!=im->rgb_height)\
   ? Imlib_clone_scaled_image(id,im,width,height)\
   : im )


/* search for an image in the configured paths and scale it
 */
ImlibImage* getImage(ImlibData *id,char *image,
		     int width,int height) {
  ImlibImage *im=NULL;
  pushArgs(256+(image?strlen(image):0),
	   (BUF,"%s,%s,%d,%d",
	    PTR2STR(id),image?image:"NULL",
	    width,height));

  if (!id || !image)
    return NULL;

  if (*image=='~') {
    char *home=getenv("HOME");
    if (home) {
      int len=strlen(home);
      char buffer[len+strlen(image)+1]; /* +1: \0 */

      trace( strcpy(buffer,home) );
      trace( buffer[len]='/' );
      trace( strcpy(&buffer[len+1],image+1) );
      trace( im=Imlib_load_image(id,buffer) );
    }
  } else if (*image!='/') {
    char buffer[strlen(image)+2+longest_path_size];
    int i;
    assert((npaths==0)==(icon_paths==NULL),"Inconsistent data");
    for(i=0;i<npaths && im==NULL;i++) {
      int len;
      assert(icon_paths[i]!=NULL,"Null path pointer");
      trace( len=strlen(icon_paths[i]) );
      assert(longest_path_size>=len,"Inconsistent longest path");
      trace( strcpy(buffer,icon_paths[i]) );
      trace( buffer[len]='/' );
      trace( strcpy(&buffer[len+1],image) );
      trace( im=Imlib_load_image(id,buffer) );
    }
  }
  if (im==NULL) /* last chance */
    trace( im=Imlib_load_image(id,image) );

  if (im==NULL)
    return NULL;
  else if (width<0 || height<0)
    return im;
  else {
    ImlibImage *scaled;
    trace( scaled=scale_image(id,im,width,height) );
    if (scaled!=im)
      trace( Imlib_kill_image(id,im) );
    return scaled;
  }
}


void printStringImage(int x,int y,ImlibImage *im,char *str,
		      FontInfo *font,
		      int* gradCol) {
  unsigned char *p;
  int width,height,len,i,jumpw,h,fw;
  int grad[3];

  pushArgs(256+STRSIZE(str),(BUF,"%d,%d,%s,%s,%s,%s",
			     x,y,PTR2STR(im),str,PTR2STR(font),
			     PTR2STR(gradCol)));
  assert(im!=NULL,"Null image");
  assert(str!=NULL,"Null string");
  assert(x>=0,"Bad value");
  assert(y>=0,"Bad value");
  
  width=im->rgb_width;
  height=im->rgb_height;
  len=(width-x)/font->width;
  h=font->height;
  if (y+h>height-1)
    h=height-y-1;
  width*=3;
  fw=font->width;
  jumpw=3*fw;
  p=im->rgb_data+3*x+y*width;
  
  for(i=0;i<3;i++)
    grad[i]=gradCol[2*i+1]-gradCol[2*i];

  if (gradCol[0]>=0) /* use color */
    for(i=0;i<len && str[i] && str[i]!='\n';i++,p+=jumpw) {
      unsigned char *ptr=p;
      if (str[i]-FIRST_CHAR>=0 && str[i]-FIRST_CHAR<NB_CHARS) {
	unsigned int *data=font->data+(str[i]-FIRST_CHAR)*font->size;
	int j,count;
	unsigned int code=0;
	debugPrint(("%c",str[i]));
	for(j=0,count=0;j<h;j++,ptr+=width) {
	  int col;
	  for(col=0;col<fw;col++,count++,code>>=1) {
	    if ((count&31)==0)
	      code=*data++;
	    if (code&1) {
	      int k;
	      for(k=0;k<3;k++) {
		trace(ptr[3*col+k]=gradCol[2*k]+(j*grad[k])/(h-1));
		trace(ptr[3*col+3+width+k]>>=1);
	      }
	    }
	  }
	}
      }
    }
  else /* use shading */
    for(i=0;i<len && str[i] && str[i]!='\n';i++,p+=jumpw) {
      unsigned char *ptr=p;
      if (str[i]-FIRST_CHAR>=0 && str[i]-FIRST_CHAR<NB_CHARS) {
	unsigned int *data=font->data+(str[i]-FIRST_CHAR)*font->size;
	int j,count;
	unsigned int code=0;
	debugPrint(("%c",str[i]));
	for(j=0,count=0;j<h;j++,ptr+=width) {
	  int col;
	  for(col=0;col<fw;col++,count++,code>>=1) {
	    if ((count&31)==0)
	      code=*data++;
	    if (code&1) {
	      int k;
	      for(k=0;k<3;k++) {
		unsigned int pix=(ptr[3*col+k]+8)<<1;
		ptr[3*col+k]=(pix<32?32:(pix>255?255:pix));
		trace(ptr[3*col+3+width+k]>>=1);
	      }
	    }
	  }
	}
      }
    }

  debugPrint(("\n"));
}


void putImlibImage(ImlibData *id,
		   ImlibImage *dst,ImlibImage *src,
		   int x,int y) {
  ImlibColor imcol;
  int width,height,i,nd,ns;
  unsigned char *s,*d;

  pushArgs(256,(BUF,"id=%s,dst=%s,src=%s,x=%d,y=%d",
		PTR2STR(id),PTR2STR(dst),PTR2STR(src),x,y));

  assert(id!=NULL,"Null image data pointer");
  assert(dst!=NULL,"Null destination pointer");
  assert(src!=NULL,"Null source pointer");
  debugPrint(("putImlibImage(%s[%dx%d],%s[%dx%d],%d,%d)\n",
	      dst->filename,dst->rgb_width,dst->rgb_height,
	      src->filename,src->rgb_width,src->rgb_height,x,y));

  trace( Imlib_get_image_shape(id,src,&imcol) );
  width=src->rgb_width;
  height=src->rgb_height;
  ns=3*width;
  nd=3*dst->rgb_width;
  s=src->rgb_data;
  d=dst->rgb_data+3*x+y*nd;
  if (x+width>dst->rgb_width)
    width=dst->rgb_width-x;
  if (y+height>dst->rgb_height)
    height=dst->rgb_height-y;

  for(i=0;i<height;i++,s+=ns,d+=nd) {
    int j;
    for(j=0;j<3*width;j+=3)
      if (!(imcol.r==s[j] &&
	    imcol.g==s[j+1] &&
	    imcol.b==s[j+2])) {
	trace(d[j]=s[j]);
	trace(d[j+1]=s[j+1]);
	trace(d[j+2]=s[j+2]);
      }
  }
}


void copyImlibImage(ImlibData *id,
		    ImlibImage *dst,ImlibImage *src,
		    int blend) {
  ImlibColor imcol;
  int width,height,i,nd,ns;
  unsigned char *s,*d;

  pushArgs(256,(BUF,"id=%s,dst=%s,src=%s,blend=%d",
		PTR2STR(id),PTR2STR(dst),PTR2STR(src),blend));

  assert(id!=NULL,"Null image data pointer");
  assert(dst!=NULL,"Null destination pointer");
  assert(src!=NULL,"Null source pointer");

  width=dst->rgb_width;
  height=dst->rgb_height;

  assert(width<=src->rgb_width,"Destination width smaller than source width!");
  assert(height<=src->rgb_height,"Destination height smaller than source height!");

  trace( Imlib_get_image_shape(id,src,&imcol) );

  d=dst->rgb_data;
  s=src->rgb_data;
  nd=3*width;
  ns=3*src->rgb_width;

  for(i=0;i<height;i++,d+=nd,s+=ns) {
    int j;
    for(j=0;j<nd;j+=3) {
      if (!(imcol.r==s[j] && imcol.g==s[j+1] && imcol.b==s[j+2])) {
	int k;
	for(k=0;k<3;k++) {
	  int tmp;
	  trace( tmp=(((int)s[j+k])-((int)d[j+k]))*blend );
	  trace( d[j+k]=((int)d[j+k]) + ((tmp + (tmp >> 8) + 0x80) >> 8) );
	}
      }
    }
  }
}



