/* utils.c
 *
 * Part of FancyLauncher
 * author: Alexandre David
 * e-mail: adavid@docs.uu.se
 * license: GPL
 */

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "utils.h"
#include "debug.h"


/* equivalent to strdup, but returns a RefString ;
 * reference counter initialized to 1
 */
RefString strdup2ref(const char *str) {
  RefString s;

  pushArgs(str?strlen(str)+1:8,(BUF,"%s",str));
  assert(str!=NULL,"Null string argument");

  trace( s=(RefString)safeAllocPtr(alloc(strlen(str)+sizeof(int)+1)) );
  REF(s)=1;
  trace( strcpy(CHARS(s),str) );

  debugPrint(("<+:"GREEN"%d"BLACK"> %s\n",REF(s),CHARS(s)));
  return s;
}


/* check pointer to be returned
 */
void* safeAllocPtr(void *p) {
  pushArgs(8,(BUF,"%s",PTR2STR(p)));
  if (!p) {
    fprintf(stderr,"Virtual memory exhausted, aborting!\n");
    exit(1);
  }
  return p;
}


/* jump blanks and return TRUE if
 * the remaining string is not empty
 * *index is updated
 */
int jumpBlank(const char *buf,int *index) {
  pushArgs(16,(BUF,"%s,%s",PTR2STR(buf),PTR2STR(index)));
  assert(buf!=NULL,"Null string argument");
  assert(index!=NULL,"Null index pointer");
  assert(index==NULL || *index>=0,"Bad index value");
  debugPrint(("jumpBlank>"));

  trace( while(buf[*index] && isBlank(buf[*index])) {
    debugPrint((GREEN"%c"BLACK,buf[*index]));
    (*index)++;
  });

  trace( debugPrint(("%s",buf+*index)) );
  return buf[*index];
}


/* return a pointer after the word
 * to jump ; write \0 after the jumped word
 */
char *jumpWordNull(char *buf) {
  pushArgs(8,(BUF,"%s",PTR2STR(buf)));
  assert(buf!=NULL,"Null string pointer");
  debugPrint(("jumpWordNull>"));

  trace( while(*buf && !isBlank(*buf)) {
    debugPrint((GREEN"%c"BLACK,*buf));
    buf++;
  });

  trace( if (*buf) {
    debugPrint((RED" "BLACK));
    *buf++=0;
  });

  debugPrint(("\n"));
  return buf;
}



/* read a token and return its code
 * *index is updated
 */
int readToken(const char *buf,int *index) {
  int code=0;

  pushArgs(16,(BUF,"%s,%s",PTR2STR(buf),PTR2STR(index)));
  assert(buf!=NULL,"Null string pointer");
  assert(index!=NULL,"Null index pointer");
  assert(index==NULL || *index>=0,"Bad index value");
  debugPrint(("readToken>"));

  for(;;) {
    int c;
    trace( c=tolower(buf[*index]) );
    if (c==0 || isBlank(c)) {
      debugPrint((" : 0x%x\n",code));
      return code;
    }
    debugPrint((GREEN"%c"BLACK,buf[*index]));
    code=37*(code+c);
    (*index)++;
  }
}


/* read an integer
 * *index is updated
 */
int readInt(char *buf,int *index) {
  int result;

  pushArgs(16,(BUF,"%s,%s",PTR2STR(buf),PTR2STR(index)));
  assert(buf!=NULL,"Null string pointer");
  assert(index!=NULL,"Null index pointer");
  assert(index==NULL || *index>=0,"Bad index value");

  trace( buf+=*index );
  trace( *index+=jumpWordNull(buf)-buf );
  debugPrint(("readInt>%s\n",buf));
  if (!strncmp(buf,"0x",2))
    sscanf(buf+2,"%x",&result);
  else
    sscanf(buf,"%d",&result);
  return result;
}


/* read a string ; return a RefString
 */
RefString readString(char *buf, int *index) {
  RefString res;

  pushArgs(16,(BUF,"%s,%s",PTR2STR(buf),PTR2STR(index)));
  assert(buf!=NULL,"Null string pointer");
  assert(index!=NULL,"Null index pointer");
  assert(index==NULL || *index>=0,"Bad index value");

  trace( buf+=*index );
  trace( *index+=jumpWordNull(buf)-buf );
  debugPrint(("readString>%s\n",buf));
  trace( res=strdup2ref(buf) );

  return res;
}


/* read a string ; return a char*, the whole remaining line is returned
 * the string is written at char* +offset
 */
char * readChars(char *buf, int *index,const int offset) {
  char *s,*arg;
  int len;

  pushArgs(64,(BUF,"%s,%s,%d",PTR2STR(buf),PTR2STR(index),offset));
  assert(buf!=NULL,"Null string pointer");
  assert(index!=NULL,"Null index pointer");
  assert(index==NULL || *index>=0,"Bad index value");

  trace( arg=buf+*index );
  trace( len=strlen(arg) );
  trace( (*index)+=len );
  trace( while(len>0 && isBlank(arg[len-1])) /* strip */
	 arg[--len]=0 );

  debugPrint(("readChars>"GREEN"%s"BLACK"\n",arg));

  trace( s=(char*)safeAllocPtr(alloc(len+1+offset)) );
  trace( strcpy(s+offset,arg) );

  return s;
}


/* check a directory and create it if it does not exist
 * handles simple directory, not paths
 */
void checkDirectory(const char *dir) {
  struct stat dir_stat;

  pushArgs(dir?strlen(dir)+1:8,(BUF,"%s",dir));
  assert(dir!=NULL,"Null string pointer");
  debugPrint(("checkDirectory(%s) ... ",dir));

  if (stat(dir,&dir_stat)) {
    perror(dir);
    fprintf(stderr,"Attempting to create directory %s ... ",dir);
    if (mkdir(dir,0755)) {
      perror("mkdir");
      exit(1);
    } else
      fprintf(stderr,"success\n");
  } else {
    if (!S_ISDIR(dir_stat.st_mode)) {
      fprintf(stderr,"%s is not a directory, remove the file and try again.\n",dir);
      exit(1);
    }
    debugOK();
  }
}


/* find code for a given key definition
 * return 0 if not found
 * open /usr/include/X11/keysymdef.h and look for the symbol
 */
int findCode(const char *short_symbol) {
  FILE *f=fopen("/usr/include/X11/keysymdef.h","r");
  char buf[1024];
  char symbol[strlen(short_symbol)+4];
  
  pushArgs(short_symbol?strlen(short_symbol)+1:8,(BUF,"%s",short_symbol));

  if (!f) {
    perror("/usr/include/X11/keysymdef.h");
    return 0;
  }
  assert(short_symbol!=NULL,"Symbol to find is null");

  symbol[0]='X';
  symbol[1]='K';
  symbol[2]='_';
  strcpy(&symbol[3],short_symbol);

  printDisable();

  while(fgets(buf,sizeof(buf),f)) { /* load each line */
    if (!strncmp(buf,"#define ",8)) { /* if #define */
      int i=8;
      if (jumpBlank(buf,&i)) { /* go to symbol definition */
	char *end=jumpWordNull(buf+i); /* mark end */
	if (!strcmp(buf+i,symbol)) { /* check right symbol */
	  i+=end-&buf[i];
	  if (jumpBlank(buf,&i)) { /* go to code */
	    jumpWordNull(buf+i); /* mark end */
	    if (!strncmp("0x",buf+i,2))
	      sscanf(buf+i+2,"%x",&i);
	    else
	      i=atoi(buf+i);
	    fclose(f);
	    printEnable();
	    return i;
	  }
	}
      }
    }
  }
  /* not found */
  fclose(f);
  printDisable();
  return 0;
}


/* inverse of findCode: find the symbol that defines a code
 */
char* findSymbol(const int code) {
  FILE *f=fopen("/usr/include/X11/keysymdef.h","r");
  char buf[1024];

  pushMe();

  if (!f) {
    perror("/usr/include/X11/keysymdef.h");
    return NULL;
  }

  printDisable();

  while(fgets(buf,sizeof(buf),f)) { /* load each line */
    if (!strncmp(buf,"#define ",8)) { /* if #define */
      int i=8;
      if (jumpBlank(buf,&i)) { /* go to symbol definition */
	char *symbol=buf+i;
	if (!strncmp(symbol,"XK_",3)) {
	  trace(i+=jumpWordNull(symbol)-symbol); /* mark end */
	  if (jumpBlank(buf,&i)) { /* go to code */
	    trace(jumpWordNull(buf+i)); /* mark end */
	    if (!strncmp("0x",buf+i,2))
	      sscanf(buf+i+2,"%x",&i);
	    else
	      i=atoi(buf+i);
	    if (i==code) { /* found code !! */
	      fclose(f);
	      symbol=strdup(symbol+3);
	      addMemoryBlock(symbol,strlen(symbol)+1);
	      printEnable();
	      return symbol;
	    }
	  }
	}
      }
    }
  }
  /* not found */
  fclose(f);
  printEnable();
  return NULL;
}


/* utilities to help parsing */

int checkPercent(const int p,int line,int col) {
  if (p<0 || p>100) {
    fprintf(stderr,
	    "Integer(%d) at line %d, column %d "
	    "is a percentage and must be in the range 0-100\n",
	    p,line,col);
    return 50;
  }
  return p;
}


/* read an integer, check argument
 */
int readIntArg(char *buffer,int *col,const int line,int *arg) {
  pushArgs(64,(BUF,"%s,%s,%d,%s",PTR2STR(buffer),PTR2STR(col),line,PTR2STR(arg)));
  assert(buffer!=NULL,"read from null buffer");
  assert(col!=NULL,"column reference is null");
  assert(*col>=0,"bad column value");
  assert(line>0,"bad line value");
  assert(arg!=NULL,"arg reference is null");

  trace(if (checkArg(buffer,col,line,"Integer")) {
    *arg=readInt(buffer,col);
    popCall();
    return 1;
  });

  return 0;
}

/* read a double, check argument
 */
void readDoubleArg(char *buffer,int *col, const int line,double *arg) {
  pushArgs(64,(BUF,"%s,%s,%d,%s",
	       PTR2STR(buffer),PTR2STR(col),line,PTR2STR(arg)));
  assert(buffer!=NULL,"read from null buffer");
  assert(col!=NULL,"column reference is null");
  assert(*col>=0,"bad column value");
  assert(line>0,"bad line value");
  assert(arg!=NULL,"arg reference is null");

  trace(if (checkArg(buffer,col,line,"Double")) {
    buffer+=*col;
    *col+=jumpWordNull(buffer)-buffer;
    *arg=atof(buffer);
  });
}


/* check an argument ; return TRUE if OK ; jump blanks
 */
int checkArg(const char *buffer, int *col,const int line,const char *typeArg) {
  pushArgs(64,(BUF,"%s,%s,%d,%s",
	       PTR2STR(buffer),PTR2STR(col),line,PTR2STR(typeArg)));
  assert(buffer!=NULL,"read from null buffer");
  assert(col!=NULL,"column reference is null");
  assert(*col>=0,"bad column value");
  assert(line>0,"bad line value");
  assert(typeArg!=NULL,"type pointer is null");

  if (!jumpBlank(buffer,col)) {
    fprintf(stderr,debug(YELLOW)
	    "%s expected at line %d, column %d\n"
	    debug(BLACK),typeArg,line,*col+1);
    return 0;
  }
  return 1;
}


/* check that the path is home or not ; disalloc path if necessary
 */
char* checkHomePath(char *path) {
  if (path && path[0]=='~' && path[1]=='/') {
    char *home=getenv("HOME");
    if (home) {
      char *s;
      trace(s=(char*)alloc(strlen(home)+strlen(path)));
      strcpy(s,home);
      strcat(s,path+1);
      disalloc(path);
      return s;
    }
  }
  return path;
}
