/* utils.h
 * provides:
 *  type RefString
 *  memory management
 *  simple parsing utilities
 *
 * Part of FancyLauncher
 * author: Alexandre David
 * e-mail: adavid@docs.uu.se
 * license: GPL
 */

#ifndef UTILS_H
#define UTILS_H

#include <sys/stat.h>
#include <unistd.h>


/* Reference string:
 * string with reference  counter integrated ; I
 * do not use another structure with a counter
 * and a pointer to char* because this is 2
 * pointers and we need only one to
 * |sizeof(int)=counter|char[]|
 */
typedef char *RefString;


/* access to reference counter
 */
#define REF(refstr) (*((int*)(refstr)))


/* access to the string itself
 */
#define CHARS(refstr) ((refstr)+sizeof(int))
#define GETCHARS(refstr) ( refstr ? CHARS(refstr) : NULL )


/* assignment of RefString to RefString ; assignment may be NULL
 */
#define SET_REF_STRING(dst,src)\
do {\
  dst = src ;\
  if (dst) {\
    assert( REF(dst) > 0 , "Reference <=0");\
    trace( REF(dst)++ );\
    debugPrint(("<+:"GREEN"%d"BLACK"> %s\n",REF(dst),CHARS(dst)));\
  }\
} while(0)


/* free: decrement the reference or free the RefString
 */
#define FREE_REF_STRING(refstr)\
do {\
  assert( refstr==NULL || REF(refstr)>0 , "Inconsistent reference (free ref <=0)");\
  debug( if (refstr) printf("<-:"GREEN"%d"BLACK"> %s\n",REF(refstr)-1,CHARS(refstr)) );\
  trace( if (refstr && --REF(refstr)==0) trace( disalloc(refstr) ));\
} while(0)


/* avoid to free NULL pointers
 */
#define FREE(ptr) do { if (ptr) disalloc(ptr); } while(0)


/* blank character definition
 */
#define isBlank(C) \
  ((C)==' ' || (C)=='\n' || (C)=='\t' || (C)=='\r')





/* equivalent to strdup, but returns a RefString ;
 * reference counter initialized to 1
 */
RefString strdup2ref(const char *str);

/* check pointer to be returned
 */
void* safeAllocPtr(void *p);




/* jump blanks and return TRUE if
 * the remaining string is not empty
 * *index is updated
 */
int jumpBlank(const char *buf,int *index);


/* return a pointer after the word
 * to jump ; write \0 after the jumped word
 */
char *jumpWordNull(char *buf);


/* read a token and return its code
 * *index is updated
 */
int readToken(const char *buf,int *index);


/* read an integer
 * *index is updated
 */
int readInt(char *buf,int *index);


/* read a string ; return a referenced string
 */
RefString readString(char *buf, int *index);


/* read a string ; return a char*, the whole remaining line is returned
 * the string is written at char* +offset
 */
char * readChars(char *buf, int *index,const  int offset);


/* check a directory and create it if it does not exist
 */
void checkDirectory(const char *dir);



/* find code for a given key definition
 * return 0 if not found
 * open /usr/include/X11/keysymdef.h and look for the symbol
 */
int findCode(const char *short_symbol);


/* inverse of findCode: find the symbol that defines a code
 */
char* findSymbol(const int code);


/* check condition and print error if necessary
 */
#define checkState(goodState,tokenForgotten) \
  do {\
    if (state!=goodState)\
      fprintf(stderr,"\""tokenForgotten"\" expected at line %d, column %d\n",line,col+1);\
  } while(0)


int checkPercent(const int p,int line,int col);


/* read an integer, check argument
 */
int readIntArg(char *buffer,int *col,const int line,int *arg);

/* read a double, check argument
 */
void readDoubleArg(char *buffer,int *col, const int line,double *arg);


#define CHECK_COND(cond,s) \
 (!(cond)) { fprintf(stderr,debug(YELLOW)\
	             "Unexpected \"%s\" at line %d, column %d\n"\
		     debug(BLACK),s,line,col+1); } else


/* check an argument ; return TRUE if OK ; jump blanks
 */
int checkArg(const char *buffer, int *col,const int line,const char *typeArg);


/* check that the path is home or not ; disalloc path if necessary
 */
char *checkHomePath(char *path);

#endif



